# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Network module containing Network facilitators and analysers.
"""
import datetime
import itertools
import json
import os
import re
import signal
import sys
import threading
import time
import traceback
import urllib.parse
import uuid
from collections import defaultdict
from tempfile import mkdtemp
from threading import Event, Lock
from typing import Optional

import isodate

from .basenoderun import BaseNodeRun
from .inputoutputrun import OutputRun
from .job import JobState
from .linkrun import LinkRun
from .networkanalyzer import DefaultNetworkAnalyzer
from .networkchunker import DefaultNetworkChunker
from .sinknoderun import SinkNodeRun
from .sourcenoderun import ConstantNodeRun, SourceNodeRun
from .. import exceptions, resources
from ..abc.serializable import Serializable
from ..globals import set_current_run
from ..core import vfs_plugin
from ..core.version import Version
from ..data import url
from ..datatypes import types
from ..helpers import config, iohelpers, log
from ..helpers.events import emit_event, EventType
from ..helpers.lockfile import DirectoryLock

__all__ = ['NetworkRun']


class NetworkRun(Serializable):
    """
    The Network class represents a workflow. This includes all Nodes (including ConstantNodes, SourceNodes and Sinks) and
    Links.
    """

    NETWORK_DUMP_FILE_NAME = '__fastr_network__.json'
    SOURCE_DUMP_FILE_NAME = '__source_data__.pickle.gz'
    SINK_DUMP_FILE_NAME = '__sink_data__.json'

    def __init__(self, network):
        """
        Create a new, empty Network

        :param str name: name of the Network
        :return: newly created Network
        :raises OSError: if the tmp mount in the config is not a writable directory
        """
        # Parent for when the run is part of another run (e.g. MacroNodes are involved)
        self.parent = None
        self.thread = None

        # Set result
        self.result = None

        # Basic information about the network the Run is based on, and a copy of the network state
        self._network = network.__getstate__()
        self.network_id = network.id
        self.network_version = network.version
        self.network_namespace = network.namespace
        self.network_filename = network.filename

        network_scope = network.filename

        if network_scope is not None:
            network_scope = os.path.dirname(network_scope)

            try:
                network_scope = vfs_plugin.path_to_url(network_scope)
            except exceptions.FastrMountUnknownError:
                log.warning('Network scope is not in any known mounts, cannot transfer scope to execution nodes!')
                network_scope = None

        log.info(f'Detected network scope: {network_scope}')
        self.network_scope = network_scope
        self.description = network.description

        # Info about the execution run
        self.tracking_id: Optional[str] = None
        self.timestamp: datetime.datetime = datetime.datetime.now()
        self.uuid: uuid.UUID = uuid.uuid1()

        # Preferred types
        self.preferred_types = list(network.preferred_types)

        # Recreate nodes
        self.nodelist = {k: BaseNodeRun.NODE_RUN_MAP.get(type(v).__name__)(v, parent=self) for k, v in network.nodelist.items()}

        # Set self as parent
        for node in self.nodelist.values():
            node.parent = self

        # Sink results
        self.sink_results = {}

        # Create the link list, make sure all Nodes are in place first
        self.linklist = {}
        for link_id, link in list(network.linklist.items()):
            state = link.__getstate__()

            # Change the link state to target the contents of this NetworkRun
            #  instead of the Network
            state['source'] = state['source'].replace(network.fullid, self.fullid)
            state['target'] = state['target'].replace(network.fullid, self.fullid)

            # Recreate the link inside the NetworkRun
            self.linklist[link_id] = LinkRun.createobj(state, self)

        # Copy step ids
        self.stepids = {}
        for key, nodes in network.stepids.items():
            self.stepids[key] = [self.nodelist[x.id] for x in nodes]

        self.callback_edit_lock = Lock()
        self.executing = False
        self.execution_lock = Event()
        self.execution_lock.set()

        # Check if temp dir exists, if not try to create it
        if not os.path.exists(config.mounts['tmp']):
            log.info("fast temporary directory does not exists, creating it...")
            try:
                os.mkdir(config.mounts['tmp'])
            except OSError:
                message = "Could not create fastr temporary directory ({})".format(config.mounts['tmp'])
                log.critical(message)
                raise exceptions.FastrOSError(message)

        self.tmpdir = None
        self.tmpurl = None
        self.tmpdir_lock = None

    def _get(self, path):
        value = self
        path = path.split('/')
        for part in path:
            if hasattr(value, part):
                value = getattr(value, part)
            elif hasattr(value, '__getitem__'):
                if isinstance(value, (list, tuple, OutputRun)):
                    value = value[int(part)]
                else:
                    value = value[part]
            else:
                raise exceptions.FastrLookupError('Could not find {} in {}'.format(part, value))

        return value

    def __repr__(self):
        return '<NetworkRun {})>'.format(self.id)

    def __bool__(self):
        """
        A network run is True if it finish running successfully and False otherwise
        """
        return bool(self.result)

    def __eq__(self, other):
        """
        Compare two Networks and see if they are equal.

        :param other:
        :type other: :py:class:`Network <fastr.planning.network.Network>`
        :return: flag indicating that the Networks are the same
        :rtype: bool
        """
        if not isinstance(other, NetworkRun):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['callback_edit_lock']
        del dict_self['execution_lock']
        del dict_self['_job_finished_callback']
        del dict_self['_job_status_callback']
        del dict_self['sink_results']

        dict_other = dict(vars(other))
        del dict_other['callback_edit_lock']
        del dict_other['execution_lock']
        del dict_other['_job_finished_callback']
        del dict_other['_job_status_callback']
        del dict_other['sink_results']

        return dict_self == dict_other

    def __ne__(self, other):
        """
        Tests for non-equality, this is the negated version __eq__
        """
        return not (self.__eq__(other))

    # Retrieve a Node/Link/Input/Output in the network based on the fullid
    def __getitem__(self, item):
        """
        Get an item by its fullid. The fullid can point to a link, node, input, output or even subinput/suboutput.

        :param str,unicode item: fullid of the item to retrieve
        :return: the requested item
        """
        if not isinstance(item, str):
            raise exceptions.FastrTypeError('Key should be a fullid string, found a {}'.format(type(item).__name__))

        parsed = urllib.parse.urlparse(item)
        if parsed.scheme != 'fastr':
            raise exceptions.FastrValueError('Item should be an URL with the fastr:// scheme (Found {} in {})'.format(parsed.scheme, item))

        path = parsed.path.split('/')[1:]

        if (len(path) < 2 or
                path[0] != 'networks' or
                path[1] != self.network_id or
                path[2] != str(self.network_version) or
                path[3] != 'runs' or
                path[4] != self.id):
            raise exceptions.FastrValueError(
                'URL {} does not point to anything in this network run, {}'.format(item, path)
            )

        value = self

        for part in path[5:]:
            tb_string = ''

            if hasattr(value, '__getitem__'):
                try:
                    if isinstance(value, (list, tuple, OutputRun)):
                        value = value[int(part)]
                    else:
                        value = value[part]
                except (KeyError, IndexError, TypeError, ValueError):
                    _, _, exc_traceback = sys.exc_info()
                    tb_string = traceback.format_tb(exc_traceback)
                else:
                    continue

            if hasattr(value, part):
                value = getattr(value, part)
            else:
                message = f'Could not find {part} in {value} (path {path})'
                log.critical(message)

                if tb_string:
                    log.critical(f'Original __getitem__ traceback:')
                    log.critical('\n'.join(tb_string))
                else:
                    log.critical('No __getitem__ available')

                raise exceptions.FastrLookupError(message)

        return value

    def __getstate__(self):
        """
        Retrieve the state of the Network

        :return: the state of the object
        :rtype dict:
        """
        state = {
            'timestamp': datetime.datetime.strftime(self.timestamp, '%Y-%m-%dT%H:%M:%S.%f'),
            'result': self.result,
            'uuid': str(self.uuid),
            'network_id': self.network_id,
            'network_version': str(self.network_version),
            'network_filename': self.network_filename,
            'nodelist': [x.__getstate__() for x in self.nodelist.values()],
            'linklist': [x.__getstate__() for x in self.linklist.values()],
            'preferred_types': [x.id for x in self.preferred_types],
            'tmpdir': self.tmpdir,
            'tmpurl': self.tmpurl,
            'network': self._network,
        }

        return state

    def __setstate__(self, state):
        """
        Set the state of the Network by the given state. This completely
        overwrites the old state!

        :param dict state: The state to populate the object with
        :return: None
        """
        # Basic info
        self.network = state.pop('_network', None)
        self.network_id = state.pop('network_id')
        self.network_version = Version(state.pop('network_version'))
        self.network_filename = state.pop('network_filename')
        self.parent = None

        # Info about the execution run (should be moved to NetworkRun in the future)
        self.timestamp = datetime.datetime.strptime(state.pop('timestamp'), '%Y-%m-%dT%H:%M:%S.%f')
        self.uuid = uuid.UUID(state.pop('uuid'))
        self.tmpdir = state.pop('tmpdir', None)
        self.tmpurl = state.pop('tmpurl', None)

        # Initialize nodelists empty to avoid errors further on
        self.nodelist = {}
        self.linklist = {}
        self.preferred_types = []

        # Create required locks
        self.callback_edit_lock = Lock()
        self.execution_lock = Event()
        self.execution_lock.set()
        self.executing = False

        # These should not be shared between Networks
        state.pop('callback_edit_lock', None)
        state.pop('execution_lock', None)

        # Recreate nodes
        if 'nodelist' in state:
            for node_state in state['nodelist']:
                # Get the node class
                node_class = node_state.get('class', 'NodeRun')
                node_class = BaseNodeRun.NODE_RUN_MAP.get(node_class)

                node = node_class.deserialize(node_state, self)

                log.debug('Adding node: {}'.format(node))
                self.nodelist[node.id] = node
                node.parent = self
            del state['nodelist']

        # Add preferred types
        state['preferred_types'] = [types[x] for x in state['preferred_types']]

        # Insert empty link_list
        statelinklist = state['linklist']
        state['linklist'] = {}

        # Update the objects dict
        self.__dict__.update(state)

        # Create the link list, make sure all Nodes are in place first
        for link in statelinklist:
            self.linklist[link['id']] = LinkRun.createobj(link, self)

    def check_id(self, id_):
        """
        Check if an id for an object is valid and unused in the Network. The
        method will always returns True if it does not raise an exception.

        :param str id_: the id to check
        :return: True
        :raises FastrValueError: if the id is not correctly formatted
        :raises FastrValueError: if the id is already in use
        """

        regex = r'^\w[\w\d_]*$'
        if re.match(regex, id_) is None:
            raise exceptions.FastrValueError('An id in Fastr should follow'
                                             ' the following pattern {}'
                                             ' (found {})'.format(regex, id_))

        if id_ in self.nodelist or id_ in self.linklist:
            raise exceptions.FastrValueError('The id {} is already in use in {}!'.format(id_, self.id))

        return True

    @property
    def id(self):
        """
        The id of the Network. This is a read only property.
        """
        return '{}_{}'.format(self.network_id,
                              datetime.datetime.strftime(self.timestamp,
                                                         '%Y-%m-%dT%H-%M-%S'))

    @property
    def fullid(self):
        """
        The fullid of the Network
        """
        return 'fastr:///networks/{}/{}/runs/{}'.format(self.network_id,
                                                        self.network_version,
                                                        self.id)

    @property
    def global_id(self):
        """
        The global id of the Network, this is different for networks used in
        macronodes, as they still have parents.
        """
        if self.parent is None:
            return self.fullid
        else:
            return '{}/network_run'.format(self.parent.global_id)

    @property
    def nodegroups(self):
        """
        Give an overview of the nodegroups in the network
        """
        nodegroups = defaultdict(list)
        for node in self.nodelist.values():
            if node.nodegroup is not None:
                nodegroups[node.nodegroup].append(node)
        return nodegroups

    @property
    def long_id(self):
        if self.parent is None:
            return self.network_id
        else:
            return '{}_{}_{}'.format(self.parent.parent.network_id,
                                     self.parent.id,
                                     self.network_id)

    @property
    def network(self):
        return self._network

    @property
    def sourcelist(self):
        return {k: v for k, v in self.nodelist.items() if isinstance(v, SourceNodeRun) and not isinstance(v, ConstantNodeRun)}

    @property
    def constantlist(self):
        return {k: v for k, v in self.nodelist.items() if isinstance(v, ConstantNodeRun)}

    @property
    def sinklist(self):
        return {k: v for k, v in self.nodelist.items() if isinstance(v, SinkNodeRun)}

    def set_data(self, sourcedata, sinkdata):
        # Set the source and sink data
        for id_, source in self.sourcelist.items():
            if isinstance(source, ConstantNodeRun):
                source.set_data()
            elif id_ in sourcedata:
                source.set_data(sourcedata[id_])
            else:
                raise exceptions.FastrKeyError('Could not find source data for SourceNode {}!'.format(id_))

            source.update()

        for id_, sink in self.sinklist.items():
            if isinstance(sinkdata, str):
                sink.set_data(sinkdata)
            elif isinstance(sinkdata, dict):
                if id_ not in sinkdata:
                    raise exceptions.FastrKeyError('Could not find sink data for SinkNode {}!'.format(id_))
                sink.set_data(sinkdata[id_])
            else:
                raise exceptions.FastrTypeError('Sinkdata should be a dict or str, found a {}!'.format(
                    type(sinkdata).__name__
                ))

    def generate_jobs(self):
        log.debug('=== GENERATE JOBS FOR {} ==='.format(self.id))
        # Create execution objects
        chuncker = DefaultNetworkChunker()
        analyzer = DefaultNetworkAnalyzer()

        # Create a network chuncker to Chunk the Network in executable blocks
        chunks = chuncker.chunck_network(self)

        for chunk in chunks:
            if not self.executing:
                return

            # Create a network analyzer to create the optimal execution order
            executionlist = analyzer.analyze_network(self, chunk)

            # Do not create sink jobs for a macro node
            if self.parent is not None:
                executionlist = [x for x in executionlist if not isinstance(x, SinkNodeRun)]

            # log.warning(f"===> {executionlist=}")
            # test1 = [x.execute() for x in executionlist]
            # log.warning(f"===> {test1=}")
            # test2 = list(itertools.zip_longest(*test1))
            # log.warning(f"===> {test2=}")

            for job_lists in itertools.zip_longest(*[x.execute() for x in executionlist]):
                if not self.executing:
                    return

                # Flatten the array of arrays while dropping the Nones (drained iterators)
                job_list = (x for y in job_lists if y is not None for x in y)

                # Only try to process chunks that actually have jobs...
                yield job_list

        log.debug('=== END GENERATE JOBS FOR {} ==='.format(self.id))

    def execute(self,
                sourcedata,
                sinkdata,
                execution_plugin=None,
                tmpdir=None,
                cluster_queue=None,
                timestamp=None,
                tracking_id: Optional[str]=None):
        """
        Execute the Network with the given data. This will analyze the Network,
        create jobs and send them to the execution backend of the system.

        :param dict sourcedata: dictionary containing all data for the sources
        :param dict sinkdata: dictionary containing directives for the sinks
        :param str execution_plugin: the execution plugin to use (None will use the config value)
        :raises FastrKeyError: if a source has not corresponding key in sourcedata
        :raises FastrKeyError: if a sink has not corresponding key in sinkdata
        """
        log.debug('Acquiring execution lock...')
        self.execution_lock.wait()
        self.execution_lock.clear()
        self.executing = True
        set_current_run(self)
        self.register_signals()

        if timestamp:
            if isinstance(timestamp, str):
                timestamp = isodate.parse_datetime(timestamp)
            self.timestamp = timestamp

        if tracking_id:
            self.tracking_id = tracking_id
        else:
            self.tracking_id = None

        log.info("####################################")
        log.info("#     network execution STARTED    #")
        log.info("####################################")

        # Display the main file the Network is created from, if known it
        # is the network.filename, otherwise try to estimate the entry
        # point for the script used
        entry_file = self.network_filename

        if entry_file is None:
            for x in sys.argv[:2]:
                if os.path.isfile(x):
                    entry_file = x
                    break

        # If there is no entry_file or the first argument is ipython or
        # bpython we assume that is not a network but via prompt
        if entry_file is not None and not entry_file.endswith((os.path.sep + "ipython",
                                                               os.path.sep + "bpython")):
            entry_file = os.path.abspath(entry_file)
            entry_file_date = time.ctime(os.path.getmtime(entry_file))
            message = 'Running network via {} (last modified {})'.format(entry_file, entry_file_date)
        else:
            if entry_file is not None:
                entry_file = os.path.basename(entry_file)
            message = 'Running network via {} session'.format(entry_file or 'python')

        log.info(message)
        log.info('FASTR loaded from {}'.format(config.systemdir))

        if tmpdir is None:
            tmpdir = os.path.normpath(mkdtemp(prefix='fastr_{}_'.format(self.id),
                                              dir=config.mounts['tmp']))
        else:
            if url.isurl(tmpdir):
                if not tmpdir.startswith('vfs://'):
                    raise exceptions.FastrValueError('The tmpdir keyword argument should be a path or vfs:// url!')
                tmpdir = vfs_plugin.url_to_path(tmpdir)

            if not os.path.exists(tmpdir):
                os.makedirs(tmpdir)

        self.tmpdir = tmpdir
        self.tmpurl = vfs_plugin.path_to_url(self.tmpdir)
        log.info('Network run tmpdir: {}'.format(self.tmpdir))

        self.tmpdir = tmpdir
        self.tmpdir_lock = DirectoryLock(self.tmpdir)
        if not self.tmpdir_lock.acquire():
            log.critical('Cannot lock temporary directory {}, aborting!'.format(self.tmpdir))
            self.abort()
            return False

        network_file = os.path.join(self.tmpdir, self.NETWORK_DUMP_FILE_NAME)

        # Save a copy of the Network
        iohelpers.save_json(network_file, self.network, indent=2)

        iohelpers.save_gpickle(os.path.join(self.tmpdir,
                                            '__source_run_{}__.pickle.gz'.format(
                                                datetime.datetime.strftime(self.timestamp,
                                                                           '%Y-%m-%dT%H-%M-%S')
                                            )),
                               sourcedata)

        # Make sure all reporting plugins are loaded
        for plugin_name in config.reporting_plugins:
            resources.plugins[plugin_name].activate()

        # Signal event that run is started
        emit_event(EventType.run_started, self)

        try:
            self.set_data(sourcedata, sinkdata)

            # Select desired server execution plugin and instantiate
            log.info('Selecting {} as executor plugin'.format(config.execution_plugin))

            # Checlk what execution plugin to use
            if execution_plugin is None:
                execution_plugin = config.execution_plugin

            if execution_plugin not in resources.execution_plugins:
                message = 'Selected non-existing execution plugin ({}), available plugins: {}'.format(execution_plugin,
                                                                                                      list(resources.execution_plugins.keys()))
                log.error(message)
                raise exceptions.FastrValueError(message)

            log.debug('Retrieving execution plugin ({})'.format(execution_plugin))
            execution_interface_type = resources.execution_plugins[execution_plugin]
            log.debug('Creating execution interface')

            with execution_interface_type(self.job_finished, self.job_finished) as execution_interface:
                if cluster_queue is not None:
                    execution_interface.default_queue = cluster_queue

                for job_list in self.generate_jobs():
                    # Queue jobs but allow cancelling in between
                    for job in job_list:
                        if not self.executing:
                            return
                        log.info(f"Queueing job {job.id}")
                        execution_interface.queue_job(job)

                    log.info('Waiting for execution to finish...')
                    # Wait in chuncks of two seconds (hopefully this will allow ipython to update more regularly)
                    last_print = time.time()
                    while len(execution_interface.job_dict) > 0:
                        sys.stdout.flush()
                        if not self.executing:
                            return

                        # Print out one which jobs we are still waiting
                        if time.time() - last_print > 600:
                            jobs = execution_interface.job_dict.values()
                            log.info(f'Waiting for {len(jobs)} jobs:')
                            if len(jobs) <= 15:
                                for job in jobs:
                                    log.info(f'{job.id}: {job.status}')
                            else:
                                # Only show first and last 5, hide rest
                                jobs = list(jobs)
                                for job in jobs[:5]:
                                    log.info(f'{job.id}: {job.status}')
                                log.info('---- {} JOBS HIDDEN ----'.format(len(jobs) - 10))
                                for job in jobs[-5:]:
                                    log.info(f'{job.id}: {job.status}')
                            last_print = time.time()
                        time.sleep(1.0)

                    if not self.executing:
                        return

                    log.info('Chunk execution finished!')

            log.info("####################################")
            log.info("#    network execution FINISHED    #")
            log.info("####################################")
        finally:
            if not self.executing:
                log.info("####################################")
                log.info("#    network execution ABORTED     #")
                log.info("####################################")

            self.execution_finished()

        sink_data_json = {
            sink_node: {
                str(sample_id): {
                    "status": str(job.status),
                    "sample_status": status.name,
                    "errors": [list(str(e) for e in x) for x in failed_annotations],
                } for sample_id, (job, status, failed_annotations) in sink_data.items()
            } for sink_node, sink_data in self.sink_results.items()
        }

        sink_result_file = os.path.join(self.tmpdir, self.SINK_DUMP_FILE_NAME)
        with open(sink_result_file, 'w') as fh_out:
            json.dump(sink_data_json, fh_out, indent=2)

        return self.result

    def execution_finished(self):
        # Make sure to unlock the Network
        log.debug('Releasing execution lock')
        self.executing = False
        self.execution_lock.set()
        self.unregister_signals()

        # FIXME: This has be to move to AFTER the last write in the tmpdir
        #        has finished, but inside a safety such as a with or finally.
        if self.tmpdir_lock:
            self.tmpdir_lock.release()

        # Check if everything was successful
        result = not any(x[1].name == 'FAILED' for sink_data in self.sink_results.values() for x in sink_data.values())
        self.result = result

        # Signal event that run finished
        emit_event(EventType.run_finished, self)

    def abort(self, signal_code=None, current_frame=None):
        if signal_code is not None:
            signal_obj = signal.Signals(signal_code)
            log.info('Recieved a {!s} signal, aborting execution of NetworkRun {}'.format(signal_obj, self.id))
        else:
            log.info('Aborting execution of NetworkRun {}'.format(self.id))

        self.executing = False
        self.unregister_signals()

    def register_signals(self):
        """
        Register handles to handle SIGINT and SIGTERM handlers to
        gracefully shut down the execution
        :return:
        """
        if threading.current_thread() is not threading.main_thread():
            log.warning('Cannot change signals from outside of main thread!')
            return

        signal.signal(signal.SIGINT, self.abort)

    def unregister_signals(self):
        """
        Unregister the signal handlers (set to default). Sending these signals
        twice will result that the second time the default handler is used.
        """
        if threading.current_thread() is not threading.main_thread():
            log.warning('Cannot change signals from outside of main thread!')
            return

        signal.signal(signal.SIGINT, signal.default_int_handler)

    def job_finished(self, job):
        """
        Call-back handler for when a job is finished. Will collect the results
        and handle blocking jobs. This function is automatically called when
        the execution plugin finished a job.

        :param job: the job that finished
        :type job: :py:class:`Job <fastr.execution.job.Job>`
        """
        # Find the Node for this job
        node = self[job.node_global_id]

        # Collect earlier failed annotations
        failed_job_annotation = set()
        if not isinstance(node, SourceNodeRun):
            for input_id, input_argument in job.input_arguments.items():

                # Check if input is hidden and system-generated (it will
                # not be part of the provenance)
                input_description = node.tool.inputs[input_id]
                if input_description.hidden:
                    log.debug('Skipping hidden system input: {}'.format(input_id))
                    continue

                if not isinstance(input_argument, list):
                    input_argument = [input_argument]

                for input_argument_element in input_argument:
                    if str(input_argument_element.id) == '__EMPTY__':
                        continue

                    # Get the annotation from the input
                    log.debug('Getting annotation from {} --> {}'.format(node.inputs[input_id].get_sourced_outputs(),
                                                                         node.inputs[input_id][input_argument_element.index].failed_annotations))
                    item = node.inputs[input_id][input_argument_element.index]
                    failed_job_annotation.update(item.failed_annotations)

        # Log the status and add annotations if needed
        status_message = "Finished job {} with status {}".format(job.id, job.status)
        log.info(status_message)

        if job.status not in (JobState.finished, JobState.cancelled):
            if len(job.errors) > 0:
                error = job.errors[0]
                status_message = 'Encountered error: [{e[0]}] {e[1]} ({e[2]}:{e[3]})'.format(e=error)
            else:
                status_message = 'Encountered error in job {}'.format(job.id)

            job_relative_path = job.logurl.replace(self.tmpurl, '.')
            failed_job_annotation.add((job.id, str(job.status), status_message, job_relative_path))

        # Lock Network for editing, make sure not multiple threads will edit network data at the same time
        with self.callback_edit_lock:
            # Data is stored here, need to place it back in the Nodes
            node.set_result(job, failed_job_annotation)

        sys.stdout.flush()




"""
Module with helper for shell escaping
"""

import os
import re
import shlex

__all__ = ['quote_argument']


if os.name == 'nt':
    META_CHARS = '()%!^"<>&|'
    META_RE = re.compile('(' + '|'.join(re.escape(char) for char in list(META_CHARS)) + ')')
    META_MAP = {char: f"^{char}" for char in META_CHARS}

    def _escape_meta_chars(m):
        char = m.group(1)
        return META_MAP[char]

    def quote_argument(arg: str) -> str:
        """
        Quote and escape argument properly for use with cmd.exe

        :param arg: argument to quote
        :return: argument with quotes for safe use in a bash-like shell
        """
        # Escape the argument for the cmd.exe shell.
        # See http://blogs.msdn.com/b/twistylittlepassagesallalike/archive/2011/04/23/everyone-quotes-arguments-the-wrong-way.aspx
        #
        # First we escape the quote chars to produce a argument suitable for
        # CommandLineToArgvW. We don't need to do this for simple arguments.

        if not arg or re.search(r'(["\s])', arg):
            arg = '"' + arg.replace('"', r'\"') + '"'

        return META_RE.sub(_escape_meta_chars, arg)
else:
    def quote_argument(arg: str) -> str:
        """
        Use shlex module to quote the argument properly
        :param arg: argument to quote
        :return: argument with quotes for safe use in a bash-like shell
        """
        return shlex.quote(arg)





# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import pytest

from fastr import exceptions
from fastr.core.version import Version


@pytest.fixture(scope="module")
def version_data():
    data_container = namedtuple('DataContainer', [
        'data1',
        'data2',
        'data3',
        'data4',
        'data5',
        'data6',
        'data7',
        'data8',
    ])

    return data_container(
        Version(),
        Version(0, 0, None, 'b', 42),
        Version((1, 2)),
        Version([3, 4, [5, 6], 'beta', 3, None, '-']),
        Version('0.1-r3'),
        Version('1.2.3.4.5.6.7-beta8_with_suffix'),
        Version([3, 4, (5, 6), 'beta', 3, None, '-']),
        Version('3.3.2'),
    )


def test_data(version_data):
    assert version_data.data1 == (0, 0, None, None, None, None, None)
    assert version_data.data2 == (0, 0, None, 'b', 42, None, None)
    assert version_data.data3 == (1, 2, None, None, None, None, None)
    assert version_data.data4 == (3, 4, (5, 6), 'beta', 3, None, '-')
    assert version_data.data5 == (0, 1, None, 're', 3, '', '-')
    assert version_data.data6 == (1, 2, (3, 4, 5, 6, 7), 'beta', 8, '_with_suffix', '-')
    assert version_data.data7 == (3, 4, (5, 6), 'beta', 3, None, '-')
    assert version_data.data8 == (3, 3, (2,), None, None, '', None)


def test_string(version_data):
    assert str(version_data.data1) == '0.0'
    assert str(version_data.data2) == '0.0b42'
    assert str(version_data.data3) == '1.2'
    assert str(version_data.data4) == '3.4.5.6-beta3'
    assert str(version_data.data5) == '0.1-r3'
    assert str(version_data.data6) == '1.2.3.4.5.6.7-beta8_with_suffix'
    assert str(version_data.data7) == '3.4.5.6-beta3'
    assert str(version_data.data8) == '3.3.2'


def test_major(version_data):
    assert version_data.data1.major == 0
    assert version_data.data2.major == 0
    assert version_data.data3.major == 1
    assert version_data.data4.major == 3
    assert version_data.data5.major == 0
    assert version_data.data6.major == 1
    assert version_data.data7.major == 3
    assert version_data.data8.major == 3


def test_minor(version_data):
    assert version_data.data1.minor == 0
    assert version_data.data2.minor == 0
    assert version_data.data3.minor == 2
    assert version_data.data4.minor == 4
    assert version_data.data5.minor == 1
    assert version_data.data6.minor == 2
    assert version_data.data7.minor == 4
    assert version_data.data8.minor == 3


def test_extra(version_data):
    assert version_data.data1.extra is None
    assert version_data.data2.extra is None
    assert version_data.data3.extra is None
    assert version_data.data4.extra == (5, 6)
    assert version_data.data5.extra is None
    assert version_data.data6.extra == (3, 4, 5, 6, 7)
    assert version_data.data7.extra == (5, 6)
    assert version_data.data8.extra == (2,)


def test_extra_string(version_data):
    assert version_data.data1.extra_string == ''
    assert version_data.data2.extra_string == ''
    assert version_data.data3.extra_string == ''
    assert version_data.data4.extra_string == '.5.6'
    assert version_data.data5.extra_string == ''
    assert version_data.data6.extra_string == '.3.4.5.6.7'
    assert version_data.data7.extra_string == '.5.6'
    assert version_data.data8.extra_string == '.2'


def test_wrong_version_int():
    # Cannot use int, should raise an Error
    with pytest.raises(exceptions.FastrVersionInvalidError):
        Version(2)


def test_wrong_version_float():
    # Cannot use float, should raise an Error
    with pytest.raises(exceptions.FastrVersionInvalidError):
        Version(3.0)


def test_wrong_version_invalid_str():
    with pytest.raises(exceptions.FastrVersionInvalidError):
        Version('test123')


def test_wrong_version_list_too_long():
    with pytest.raises(exceptions.FastrVersionInvalidError):
        Version(list(range(10)))




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import OrderedDict

import sympy

from .flownoderun import FlowNodeRun
from .inputoutputrun import SourceOutputRun
from .job import SourceJob, JobState
from .. import exceptions, MISSING
from ..core import vfs_plugin as vfs
from ..core.samples import SampleId, SampleItem, SampleIndex, SampleState
from ..data import url
from ..datatypes import DataType, types
from ..helpers import log

__all__ = ['SourceNodeRun', 'ConstantNodeRun']


class SourceNodeRun(FlowNodeRun):
    """
    Class providing a connection to data resources. This can be any kind of
    file, stream, database, etc from which data can be received.
    """

    __dataschemafile__ = 'SourceNodeRun.schema.json'
    _OutputType = SourceOutputRun
    _JobType = SourceJob

    def __init__(self, node, parent):
        """ Instantiation of the SourceNodeRun.

        :param fastr.planning.node.Node node: The Node that this Run is based on.
        :param NetworkRun parent: The NetworkRun that this NodeRun belongs to
        :return: newly created sink node run
        """
        super(SourceNodeRun, self).__init__(node, parent)

        # Set the DataType
        self._input_data = None
        self._outputsize = None
        self.outputsize = 'N_{}'.format(self.id)

    def __eq__(self, other):
        """Compare two Node instances with each other. This function ignores
        the parent and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: NodeRun
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, SourceNodeRun):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_parent']
        del dict_self['_status']
        del dict_self['_input_groups']
        del dict_self['_input_data']
        del dict_self['_outputsize']
        del dict_self['input_group_combiner']

        dict_other = dict(vars(other))
        del dict_other['_parent']
        del dict_other['_status']
        del dict_other['_input_groups']
        del dict_other['_input_data']
        del dict_other['_outputsize']
        del dict_other['input_group_combiner']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the SourceNodeRun

        :return: the state of the object
        :rtype dict:
        """
        state = super(SourceNodeRun, self).__getstate__()

        return state

    def __setstate__(self, state):
        """
        Set the state of the SourceNodeRun by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(SourceNodeRun, self).__setstate__(state)

        self._input_data = None
        self._outputsize = None
        self.outputsize = 'N_{}'.format(self.id)

    @property
    def datatype(self):
        """
        The datatype of the data this source supplies.
        """
        return self.outputs['output'].datatype

    @property
    def sourcegroup(self):
        log.warning('[DEPRECATED] The sourcegroup property of the'
                    ' SourceNodeRun is deprecated and replaced by the'
                    ' nodegroup property of the NodeRun. Please use that'
                    ' property instead, it will have the same'
                    ' functionality')
        return self.nodegroup

    @property
    def dimnames(self):
        """
        Names of the dimensions in the SourceNodeRun output. These will be reflected
        in the SampleIdLists.
        """
        if self.nodegroup is not None:
            return self.nodegroup,
        else:
            return self.id,

    @property
    def output(self):
        """
        Shorthand for ``self.outputs['output']``
        """
        return self.outputs['output']

    @property
    def outputsize(self):
        """
        The size of output of this SourceNodeRun
        """
        return self._outputsize

    @outputsize.setter
    def outputsize(self, value):
        # it seems pylint does not realize this is part of a property
        # pylint: disable=arguments-differ
        if isinstance(value, str):
            self._outputsize = (sympy.symbols(value),)
        elif isinstance(value, int):
            self._outputsize = (value,)
        else:
            try:
                self._outputsize = [x if isinstance(x, int) else sympy.symbols(x.replace(' ', '_')) for x in value]
            except TypeError:
                raise exceptions.FastrTypeError('Not a valid input type')

    @property
    def valid(self):
        """
        This does nothing. It only overloads the valid method of NodeRun().
        The original is intended to check if the inputs are connected to
        some output. Since this class does not implement inputs, it is skipped.
        """
        return True

    def set_data(self, data, ids=None):
        """
        Set the data of this source node.

        :param data: the data to use
        :type data: dict, OrderedDict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        self._input_data = OrderedDict()

        # Check if data has key or generate keys
        log.debug('Storing {} (ids {}) in {}'.format(data, ids, self.fullid))
        if isinstance(data, OrderedDict):
            ids = [SampleId(x) for x in data.keys()]
            data = list(data.values())
        elif isinstance(data, dict):
            # Have data sorted on ids
            ids, data = list(zip(*sorted(data.items())))
            ids = [SampleId(x) for x in ids]
        elif isinstance(data, list):
            if ids is None:
                ids = [SampleId('id_{}'.format(k)) for k in range(len(data))]
            elif not isinstance(ids, list):
                raise exceptions.FastrTypeError('Invalid type! The ids argument should be a list that matches the data samples!')
        elif isinstance(data, tuple):
            # A single sample with cardinality
            ids = [SampleId('id_0')]
            data = [data]
        else:
            if isinstance(data, set):
                log.warning('Source data for {} is given as a set,'.format(self.fullid) +
                                  ' this is most probably a mistake and a list or dict should'
                                  ' be used instead')
            ids = [SampleId('id_0')]
            data = [data]

        log.debug('Set data in {} with {} (Type {})'.format(self.id, data, self.datatype))

        for key, value in zip(ids, data):
            if isinstance(value, tuple):
                self._input_data[key] = tuple(x if self.datatype.isinstance(x) or x is MISSING else types['String'](str(x)) for x in value)
            else:
                self._input_data[key] = (value if self.datatype.isinstance(value) or value is MISSING else types['String'](str(value))),
            log.debug('Result {}: {} (Type {})'.format(key, self._input_data[key], type(self._input_data[key]).__name__))

        self.outputsize = len(self._input_data),

    def _execute(self):
        """
        Execute the source node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        if not self.valid or self._input_data is None:
            msg = 'Cannot executed a SourceNodeRun that is not valid! Messages:\n{}'.format('\n'.join(self.messages))
            log.error(msg)
            raise exceptions.FastrValueError(msg)

        for index, (sample_id, value) in enumerate(self._input_data.items()):
            sample_index = SampleIndex(index)

            if any(x is MISSING for x in value):
                self.outputs['output'][sample_id, sample_index + (0,)] = SampleItem(sample_index + (0,),
                                                                                    sample_id,
                                                                                    {0: tuple(value)},
                                                                                    jobs=set(),
                                                                                    failed_annotations=set(),
                                                                                    status=SampleState.MISSING)

                # Broadcast fake job update?
                job = self.create_job(sample_id,
                                      sample_index,
                                      job_data={'input': value},
                                      job_dependencies=None,
                                      status=SampleState.MISSING)
                job.status = JobState.finished
            elif all(not url.isurl(str(x)) for x in value):
                # A simple string should not be send to IOPlugin for procesing
                log.debug('No job needed for sample {} at {}'.format(sample_id, self.fullid))
                self.jobs[sample_id] = None
                output_value = []

                for subvalue in value:
                    # it appears pylint does not realize that self.datatype is a class
                    # pylint: disable=not-callable
                    if self.datatype.isinstance(subvalue):
                        output_value.append(subvalue)
                    else:
                        output_value.append(self.datatype(str(subvalue)))

                self.outputs['output'][sample_id, sample_index + (0,)] = SampleItem(sample_index + (0,),
                                                                                    sample_id,
                                                                                    {0: tuple(output_value)},
                                                                                    set())

                # Broadcast fake job update?
                job = self.create_job(sample_id,
                                      sample_index,
                                      job_data={'input': value},
                                      job_dependencies=None,
                                      status=SampleState.VALID)
                job.status = JobState.finished
            else:
                # We found an URL, should be
                log.debug('Spawning job for sample {} at {}'.format(sample_id, self.fullid))
                yield self.create_job(sample_id, sample_index, {'input': value}, [], SampleState.VALID)

        self.drained = True

    def create_job(self, sample_id, sample_index, job_data, job_dependencies, status, **kwargs):
        job = super(SourceNodeRun, self).create_job(sample_id, sample_index,
                                                    job_data, job_dependencies,
                                                    status,
                                                    datatype=self.datatype.id,
                                                    **kwargs)
        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``

        .. note::
            For a SourceNodeRun this function adds a few default (hidden) arguments
        """
        log.debug('Wrapping SourceNodeRun with {}'.format(job_data))
        arguments = super(SourceNodeRun, self)._wrap_arguments(job_data, sample_id, sample_index)
        arguments[0]['input'] = SampleItem(sample_index, sample_id, job_data['input'])
        arguments[0]['datatype'] = SampleItem(sample_index, sample_id, types['String'](self.datatype.id))
        arguments[0]['sample_id'] = SampleItem(sample_index, sample_id, types['String'](str(sample_id)))

        outputurl = url.join(self.parent.tmpurl, self.id, str(sample_id), 'result')
        outputpath = vfs.url_to_path(outputurl)
        if not os.path.exists(outputpath):
            os.makedirs(outputpath)
        arguments[0]['targetdir'] = SampleItem(sample_index, sample_id, types['Directory'](outputurl))

        return arguments

    def _update(self, key, forward=True, backward=False):
        """
        Update the NodeRun information and validity of the NodeRun and propagate
         the update downstream. Updates inputs, input_groups, outputsize and outputs.

        A NodeRun is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized
        """
        # Make sure the Inputs and input groups are up to date
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        for input_ in self.inputs.values():
            input_.update(key)

        # Update own status
        valid = True
        messages = []

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('Input {} is not valid: {}'.format(id_, message))

        for input_group in self.input_groups.values():
            if input_group.empty:
                valid = False
                messages.append('InputGroup {} is empty'.format(input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None and not issubclass(output.resulting_datatype, DataType):
                valid = False
                messages.append(
                    'Output {} cannot determine the Output DataType (got {}), '
                    'please specify a valid DataType or add casts to the Links'.format(
                        id_, output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, backward)


class ConstantNodeRun(SourceNodeRun):
    """
    Class encapsulating one output for which a value can be set. For example
    used to set a scalar value to the input of a node.
    """

    __dataschemafile__ = 'ConstantNodeRun.schema.json'

    def __init__(self, node, parent):
        """
        Instantiation of the ConstantNodeRun.

        :param datatype: The datatype of the output.
        :param data: the prefilled data to use.
        :param id_: The url pattern.

        This class should never be instantiated directly (unless you know what
        you are doing). Instead create a constant using the network class like
        shown in the usage example below.

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.Network()
          >>> source = network.create_source(datatype=types['ITKImageFile'], id_='sourceN')

        or alternatively create a constant node by assigning data to an item in an InputDict:

        .. code-block:: python

          >>> node_a.inputs['in'] = ['some', 'data']

        which automatically creates and links a ConstantNodeRun to the specified Input
        """
        super(ConstantNodeRun, self).__init__(node, parent)
        self.set_data(node.data)
        self._data = node.data

    def __getstate__(self):
        """
        Retrieve the state of the ConstantNodeRun

        :return: the state of the object
        :rtype dict:
        """
        state = super(ConstantNodeRun, self).__getstate__()

        state['data'] = list(self._data.items())

        return state

    def __setstate__(self, state):
        """
        Set the state of the ConstantNodeRun by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(ConstantNodeRun, self).__setstate__(state)

        self._data = OrderedDict((SampleId(str(x) for x in key), tuple(str(x) for x in value)) for key, value in state['data'])
        self.set_data()  # Make sure that the output size etc gets set

    def set_data(self, data=None, ids=None):
        """
        Set the data of this constant node in the correct way. This is mainly
        for compatibility with the parent class SourceNodeRun

        :param data: the data to use
        :type data: dict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        # We have to arguments to match the superclas
        # pylint: disable=unused-argument
        if data is None and self.data is not None:
            self._input_data = self.data
        else:
            super(ConstantNodeRun, self).set_data(data, ids)

    @property
    def data(self):
        """
        The data stored in this constant node
        """
        return self._data

    def _execute(self):
        """
        Execute the constant node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        # Make sure the data is set
        self.set_data()

        # Run as a normal SourceNode
        for job in super(ConstantNodeRun, self)._execute():
            yield job



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import fastr
from fastr.helpers.report import print_job_result
from fastr.utils.cmd import add_parser_doc_link
from fastr.execution.job import Job
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


def get_parser():
    parser = argparse.ArgumentParser(
        description="Print a report of a job result file."
    )
    parser.add_argument('job', metavar='JOBFILE', nargs='?', type=str,
                        help=f'File of the job to execute (default ./{Job.RESULT_DUMP})')
    parser.add_argument('-v', '--verbose', action='store_true', help="More verbose (e.g. add fastr job stdout and stderr)")
    return parser


def main():
    """
    Print report of a job result (__fastr_result__.pickle.gz) file
    """
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    job = args.job

    if job is None:
        curdir_contents = os.listdir('.')

        if Job.RESULT_DUMP in curdir_contents:
            job = Job.RESULT_DUMP
        else:
            fastr.log.critical(f'No job given and cannot find {Job.RESULT_DUMP} in current directory!')
            exit()

    print_job_result(job, verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def create_network():
    # Import the faster environment and set it up
    import fastr
    # Create a new network
    network = fastr.create_network(id='collapse')

    # Create a source node in the network
    source = network.create_source('Int', id='source')

    # Create a sink to save the data
    sink = network.create_sink('Int', id='sink')

    # Link the addint node to the sink
    link = source.output >> sink.input
    link.collapse = 0

    return network


def source_data(network):
    return {
        'source': {
            'sample_a': (1, 2, 3, 4),
            'sample_b': (5, 6, 7, 8),
            'sample_c': (9, 10, 11, 12),
        }
    }


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Execute
    network.draw()
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the management of resource limits of compute resources
"""

import re
from typing import Optional, Union


class ResourceLimit:
    # Fields that are valid resource limits
    __slots__ = ('_cores', '_memory', '_time', '_gpus')
    _memory_multiplier = 1

    @classmethod
    def set_memory_multiplier(cls, value=Optional[int]) -> None:
        cls._memory_multiplier = value

    def __init__(self,
                 cores: Optional[int] = 1,
                 memory: Union[str, int, None] = '2G',
                 time: Union[str, int, None] = None,
                 gpus: Union[int, None] = None):
        """
        An object describing resource requirements/limits for a node

        :param cores: number of cores
        :param memory: memory specification, can be int with number of megabytes
                       or a string with numbers ending on M, G, T, P for megabytes,
                       gigabytes, terrabytes or petabytes. Note that the number has
                       to be an integer, e.g. 1500M would work, whereas 1.5G would
                       be invalid
        :param time: run time specification, this can be an int with the number of
                     seconds or a string in the HH:MM:SS, MM:SS, or SS format. Where
                     HH, MM, and SS are integers representing the number of hours,
                     minutes and seconds.
        :param gpus: number of GPUs
        """

        self._cores = cores
        self._memory = self._parse_memory(memory)
        self._time = self._parse_time(time)
        self._gpus = gpus

    def __eq__(self, other) -> bool:
        """
        Check if two resource limits are equal

        :param other: resource limit to test against
        """
        if not isinstance(other, ResourceLimit):
            return NotImplemented

        return (self.cores == other.cores and self.memory == other.memory
                and self.time == other.time and self.gpus == other.gpus)

    def __ne__(self, other) -> bool:
        """
        Check if two resource limits are not equal

        :param other: resource limit to test against
        """
        return not self.__eq__(other)

    def __getstate__(self) -> dict:
        return {
            "cores": self._cores,
            "memory": self._memory,
            "time": self._time,
            "gpus": self._gpus
        }

    def __setstate__(self, state: dict):
        self._cores = state.get('cores')
        self._memory = state.get('memory')
        self._time = state.get('time')
        self._gpus = state.get('gpus')

    @property
    def cores(self) -> Optional[int]:
        """
        The required number of gpus
        """
        return self._cores

    @cores.setter
    def cores(self, value: Optional[int]):
        self._cores = value

    @property
    def memory(self) -> Optional[int]:
        """
        The required memory in megabytes
        """
        return ResourceLimit._memory_multiplier * self._memory if self._memory else None

    @memory.setter
    def memory(self, value: Union[str, int, None]):
        self._memory = self._parse_memory(value) if value else None

    @staticmethod
    def _parse_memory(value: Union[str, int]) -> int:
        if value is None:
            return None
        elif isinstance(value, int):
            return value
        elif isinstance(value, str):
            match = re.match(r'^(\d+)\s*([mMgGtTpP])[Bb]?$', value)
            if match:
                number, unit = match.groups()

                # Parse number and unit
                number = int(number)
                unit = {
                    'm': 1,
                    'g': 1024,
                    't': 1024**2,
                    'p': 1024**3,
                }[unit.lower()]

                return number * unit
            else:
                raise ValueError('A memory limit string should be formatted correctly!')
        else:
            raise TypeError('Invalid type of memory limit!')

    @property
    def time(self) -> int:
        """
        The required time in seconds
        """
        return self._time if self._time else None

    @time.setter
    def time(self, value: Union[str, int, None]):
        self._time = self._parse_time(value) if value else None

    @staticmethod
    def _parse_time(value: Union[str, int]) -> int:
        if value is None:
            return None
        elif isinstance(value, int):
            return value
        elif isinstance(value, str):
            match = re.match(r'^(?:(?:(\d+):)?(\d+):)?(\d+)$', value)
            if match:
                hours, minutes, seconds = [int(x) if x else 0 for x in match.groups()]

                return hours * 3600 + minutes * 60 + seconds
            else:
                raise ValueError('A time limit string should be formatted correctly!')
        else:
            raise TypeError('Invalid type of time limit!')

    @property
    def gpus(self) -> Optional[int]:
        """
        The required number of gpus
        """
        return self._gpus

    @gpus.setter
    def gpus(self, value: Optional[int]):
        self._gpus = value

    def copy(self) -> 'ResourceLimit':
        """
        Return a copy of current resource limit object
        """
        value = ResourceLimit()
        value.__setstate__(self.__getstate__())
        return value



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Classes for arranging the input and output for nodes.

Exported classes:

Input -- An input for a node (holding datatype).
Output -- The output of a node (holding datatype and value).
ConstantOutput -- The output of a node (holding datatype and value).

.. warning::
   Don't mess with the Link, Input and Output internals from other places.
   There will be a huge chances of breaking the network functionality!
"""
import itertools
from abc import abstractmethod
from typing import Union

import sympy

from .. import exceptions
from ..core.dimension import Dimension
from ..core.samples import SampleItem, SampleId, SampleIndex, SampleValue, SampleCollection, ContainsSamples, HasSamples
from ..helpers import log
from ..planning.inputoutput import BaseInput, BaseOutput
from ..datatypes import DataType, types


class BaseInputRun(HasSamples, BaseInput):
    """
    Base class for all inputs runs.
    """

    def __init__(self, node_run, template):
        """
        Instantiate a BaseInput

        :param node: the parent node the input/output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
                            describing the input/output.
        :return: the created BaseInput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``fastr.types``
        """
        super().__init__(node_run, template.description)

    @abstractmethod
    def itersubinputs(self):
        """
        Iterator over the SubInputs

        :return: iterator

        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        raise exceptions.FastrNotImplementedError('Purposefully not implemented')


class InputRun(BaseInputRun):
    """
    Class representing an input of a node. Such an input will be connected
    to the output of another node or the output of an constant node to provide
    the input value.
    """

    def __init__(self, node_run, template):
        """
        Instantiate an input.

        :param template: the Input that the InputRun is based on
        """
        self._source = {}
        super(InputRun, self).__init__(node_run, template)
        self._input_group = template.input_group

    def __getstate__(self):
        """
        Retrieve the state of the Input

        :return: the state of the object
        :rtype dict:
        """
        state = super(InputRun, self).__getstate__()
        state['input_group'] = self.input_group

        return state

    def __setstate__(self, state):
        """
        Set the state of the Input by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(InputRun, self).__setstate__(state)
        self._input_group = state['input_group']

    def __getitem__(self, key):
        """
        Retrieve an item from this Input.

        :param key: the key of the requested item, can be a key str, sample
                    index tuple or a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :type key: str, :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` or tuple
        :return: the return value depends on the requested key. If the key was
                 an int the corresponding :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                 will be returned. If the key was a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                 or sample index tuple, the corresponding
                 :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned.
        :rtype: :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` or
                :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the key is not found
        """
        if not isinstance(key, (int, str, SampleId, SampleIndex)):
            raise exceptions.FastrTypeError('Input indices must a int, str, SampleId or '
                                            'SampleIndex, not {}'.format(type(key).__name__))

        if isinstance(key, (SampleId, SampleIndex)):
            data = []
            # Create mapping items of key; value and combine those
            self_size = self.size
            for subindex_key, sub in list(self.source.items()):
                # Allow the same mixing of parts of a mapped input as in input groups
                if sub.size == self_size:
                    value = sub[key]
                elif sub.size == (1,):
                    value = sub[SampleIndex(0)]
                elif sub.size == (0,) or sub.size == ():
                    value = SampleItem(SampleIndex(0), '__EMPTY__', SampleValue(), set(), set())
                else:
                    raise exceptions.FastrSizeMismatchError('Input has inconsistent size/dimension'
                                                            ' info for (sub)Input {}'.format(sub.fullid))

                data.append(SampleItem(value.index,
                                       value.id,
                                       {subindex_key: tuple(value.data.sequence_part())},
                                       value.jobs,
                                       value.failed_annotations,
                                       value.status))

            combination = SampleItem.combine(data)
            return combination

        if key not in self.source:
            # This is to allow for linking against inputs['key'][0]
            try:
                key = int(key)
            except ValueError:
                pass  # No problem, just go for the str

            if isinstance(key, int):
                self.source[key] = SubInputRun(self)
            else:
                self.source[key] = NamedSubinputRun(self)

        return self.source[key]

    def __str__(self):
        """
        Get a string version for the Input

        :return: the string version
        :rtype: str
        """
        return '<InputRun: {})>'.format(self.fullid)

    def get_subinput_cardinality(self, index, key=None, job_data=None):
        """
        Cardinality for a SubInput

        :param index: index for a specific sample
        :type index: int
        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """

        self_size = self.size

        sub_inputs = list(self.source.values())
        if not sub_inputs:
            # Zero arguments, so zero cardinality
            return 0
        else:
            sub_input = sub_inputs[index]

        if sub_input.size == self_size:
            cardinality = sub_input.cardinality(key, job_data)
        elif any(isinstance(x, sympy.Symbol) for x in sub_input.size):
            # This can happend during updates before the required
            # chunks are processed. Just indicate we do not know
            # the cardinality value
            return None
        elif sub_input.size == (1,):
            cardinality = sub_input.cardinality(SampleIndex(0), job_data)
        elif sub_input.size == (0,) or sub_input.size == ():
            cardinality = 0
        else:
            raise exceptions.FastrSizeMismatchError(
                f'Input has inconsistent size/dimension info for '
                f'subinput {sub_input.fullid}. Input size: {self_size}, '
                f'mismatching subinput size {sub_input.size}.'
            )

        return cardinality

    def cardinality(self, key=None, job_data=None):
        """
        Cardinality for an Input is the sum the cardinalities of the SubInputs,
        unless defined otherwise.

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """
        cardinality = 0

        for n, sub_input in enumerate(self.source.values()):
            sub_cardinality = self.get_subinput_cardinality(n, key, job_data)
            if sub_cardinality is None:
                # This can happen during updates before the required
                # chunks are processed. Just indicate we do not know
                # the cardinality value
                return None

            cardinality += sub_cardinality

        return cardinality

    def remove(self, value):
        """
        Remove a SubInput from the SubInputs list.

        :param value: the :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                      to removed from this Input
        :type value: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        """
        # Find item(s) to remove
        to_remove = []
        for key, val in self.source.items():
            if value is val:
                to_remove.append(key)

        # Remove item(s)
        for key in to_remove:
            self.source.pop(key)

    @property
    def datatype(self):
        """
        The datatype of this Input
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        # This does not differ, as it is a property
        # pylint: disable=arguments-differ
        self._datatype = value
        for subinput in self.itersubinputs():
            subinput.datatype = value

    @property
    def fullid(self):
        """
        The full defining ID for the Input
        """
        if self.node is not None:
            return '{}/inputs/{}'.format(self.node.fullid, self.id)
        else:
            return 'fastr://ORPHANED/inputs/{}'.format(self.id)

    @property
    def input_group(self):
        """
        The id of the :py:class:`InputGroup <fastr.planning.node.InputGroup>` this
        Input belongs to.
        """
        return self._input_group

    @property
    def dimensions(self):
        """
        The size of the sample collections that can accessed via this Input.
        """
        subinputs = list(self.itersubinputs())
        sizes = [sub.size for sub in subinputs]
        unique_sizes = set(sizes) - {(0,), (1,)}

        if len(unique_sizes) > 1:
            size_map = {x.source_output.id: x.size for x in self.itersubinputs()}

            # Check if the sizes can match if we ignore symbols
            for index in itertools.zip_longest(*unique_sizes):
                index = set(x for x in index if isinstance(x, int))

                if len(index) > 1:
                    message = 'Conflicting sizes of SubInputs ({}) full size map: {}'.format(unique_sizes,
                                                                                             size_map)
                    log.error(message)
                    raise exceptions.FastrSizeMismatchError(message)

            # Return dimensions of first subinput with possible size match
            for sub in subinputs:
                if sub.size in unique_sizes:
                    return sub.dimensions
        elif len(unique_sizes) == 1:
            return subinputs[sizes.index(unique_sizes.pop())].dimensions
        elif (1,) in sizes:
            return subinputs[sizes.index((1,))].dimensions
        elif (0,) in sizes:
            return subinputs[sizes.index((0,))].dimensions
        else:
            return ()

    @property
    def source(self):
        """
        The mapping of :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        that are connected and have more than 0 elements.
        """
        return self._source

    @source.setter
    def source(self, value):
        """
        The list of :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        that are connected and have more than 0 elements. (setter)
        """
        self.clear()

        self._source = {0: SubInputRun(self)}
        self._source[0].source = value

    def get_sourced_nodes(self):
        """
        Get a list of all :py:class:`Nodes <fastr.planning.node.Node>` connected as sources to this Input

        :return: list of all connected :py:class:`Nodes <fastr.planning.node.Node>`
        :rtype: list
        """
        sourced = []
        for subinput in self.itersubinputs():
            for node in subinput.get_sourced_nodes():
                if node not in sourced:
                    sourced.append(node)
        return sourced

    def get_sourced_outputs(self):
        """
        Get a list of all :py:class:`Outputs <fastr.planning.inputoutput.Output>` connected as sources to this Input

        :return: tuple of all connected :py:class:`Outputs <fastr.planning.inputoutput.Output>`
        :rtype: tuple
        """
        return tuple(n for subinput in self.itersubinputs() for n in subinput.get_sourced_outputs())

    def index(self, value):
        """
        Find index of a SubInput

        :param value: the :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                      to find the index of
        :type value: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :return: key
        :rtype: int, str
        """
        for key, val in self.source.items():
            if val is value:
                return key
        else:
            return None

    def insert(self, index):
        """
        Insert a new SubInput at index in the sources list

        :param int key: positive integer for position in _source list to insert to
        :return: newly inserted :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        """
        newsub = SubInputRun(self)
        self.source[index] = newsub
        return newsub

    def itersubinputs(self):
        """
        Iterate over the :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        in this Input.

        :return: iterator yielding  :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`


        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        for subinput in self.source.values():
            yield subinput

    def _update(self, key, forward=True, backward=False):
        """Update the validity of the Input and propagate the update downstream.
        An Input is valid if:

        * All SubInputs are valid (see :py:meth:`SubInput.update <fastr.planning.inputoutput.SubInput.update>`)
        * Cardinality is correct
        * If Input is required, it must have a size larger than (0,)
        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        for subinput in self.itersubinputs():
            subinput.update(key, forward, backward)

        valid = True
        messages = []
        for subinput in self.itersubinputs():
            if not subinput.valid:
                valid = False
                for message in subinput.messages:
                    messages.append('SubInput {} is not valid: {}'.format(subinput.fullid, message))

        if not self.check_cardinality(planning=True):
            # If the cardinality is 0 and Input is not required, this is fine,
            # all other cases where the cardinality check fails are not allowed
            valid = False
            messages.append(('Input "{}" cardinality ({}) is not valid (must'
                             ' be {}, required is {})').format(self.id,
                                                               self.cardinality(),
                                                               self._description.cardinality,
                                                               self.required))
        if self.size is None:
            valid = False
            messages.append('Cannot determine size of Input "{}"'.format(self.id))

        log.debug('Size: {}'.format(self.size))
        if self.required and (len([x for x in self.size if x != 0]) == 0):
            valid = False
            nodes = ', '.join([x.id for x in self.get_sourced_nodes()])
            messages.append(('Required Input "{}" cannot have size 0. Input obtained'
                             ' from nodes: {}').format(self.id, nodes))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update downstream
        self.node.update(key, forward, backward)


class NamedSubinputRun(InputRun):
    """
    A named subinput for cases where the value of an input is mapping.
    """
    def __init__(self, parent):
        super().__init__(parent.node, parent)
        self.parent = parent

    def __getitem__(self, key: int) -> 'Union[SubInputRun, SampleItem]':
        """
        Retrieve an item (a SubInput) from this NamedSubInput.

        :param key: the key of the requested item
        :return: The :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                 corresponding with the key will be returned.
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the key is not found
        """
        if isinstance(key, str):
            try:
                key = int(key)
            except ValueError:
                log.error(f'Could not convert key string from {key} to integer')

        if not isinstance(key, (int, SampleId, SampleIndex)):
            raise exceptions.FastrTypeError('NamedSubInputRun indices must a int, SampleId or '
                                            'SampleIndex, not {}'.format(type(key).__name__))

        if isinstance(key, (SampleId, SampleIndex)):
            data = []
            # Create mapping items of key; value and combine those
            self_size = self.size
            for subindex_key, sub in list(self.source.items()):
                # Allow the same mixing of parts of a mapped input as in input groups
                if sub.size == self_size:
                    value = sub[key]
                elif sub.size == (1,):
                    value = sub[SampleIndex(0)]
                elif sub.size == (0,) or sub.size == ():
                    value = SampleItem(SampleIndex(0), '__EMPTY__', SampleValue(), set(), set())
                else:
                    raise exceptions.FastrSizeMismatchError('Input has inconsistent size/dimension'
                                                            ' info for (sub)Input {}'.format(sub.fullid))

                data.append(SampleItem(value.index,
                                       value.id,
                                       {subindex_key: tuple(value.data.sequence_part())},
                                       value.jobs,
                                       value.failed_annotations))

            combination = SampleItem.combine(data)
            return combination

        if key not in self.source:
            self.source[key] = SubInputRun(self)

        return self.source[key]

    def __str__(self):
        """
        Get a string version for the NamedSubInput

        :return: the string version
        :rtype: str
        """
        return f'<NamedSubInput: {self.fullid})>'

    @property
    def fullid(self):
        """
        The full defining ID for the NamedSubInputRun
        """
        return '{}/{}'.format(self.parent.fullid, self.item_index)

    @property
    def item_index(self):
        return self.parent.index(self)


class SubInputRun(BaseInputRun):
    """
    This class is used by :py:class:`Input <fastr.planning.inputoutput.Input>` to
    allow for multiple links to an :py:class:`Input <fastr.planning.inputoutput.Input>`.
    The SubInput class can hold only a single Link to a (Sub)Output, but behaves
    very similar to an :py:class:`Input <fastr.planning.inputoutput.Input>` otherwise.
    """

    def __init__(self, input_):
        """
        Instantiate an SubInput.

        :param input_: the parent of this SubInput.
        :type input_: :py:class:`Input <fastr.planning.inputoutput.Input>`
        :return: the created SubInput
        """
        self._source = None

        if not isinstance(input_, InputRun):
            raise exceptions.FastrTypeError('First argument for a SubInput constructor should be an Input')

        self.parent = input_
        super(SubInputRun, self).__init__(self.node, self.parent)

        self.datatype = input_.datatype

    def __getitem__(self, key):
        """
        Retrieve an item from this SubInput.

        :param key: the key of the requested item, can be a number, sample
                    index tuple or a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :type key: int, :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` or
                   :py:class:`SampleIndex <fastr.core.sampleidlist.SampleIndex>`
        :return: the return value depends on the requested key. If the key was
                 an int the corresponding :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                 will be returned. If the key was a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                 or sample index tuple, the corresponding
                 :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned.
        :rtype: :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` or
                :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :raises FastrTypeError: if key is not of a valid type

        .. note:: As a SubInput has only one SubInput, only requesting int key
                  0 or -1 is allowed, and it will return self
        """

        if not isinstance(key, (int, SampleIndex, SampleId)):
            raise exceptions.FastrTypeError('SubInput indices must be an int, SampleIndex, or SampleID, not {}'.format(type(key).__name__))

        if isinstance(key, (SampleIndex, SampleId)):
            return self.source[0][key]

        if not -1 <= key < 1:
            raise exceptions.FastrIndexError('SubInput index out of range (key: {})'.format(key))

        return self

    def __getstate__(self):
        """
        Retrieve the state of the SubInput

        :return: the state of the object
        :rtype dict:
        """
        state = super(SubInputRun, self).__getstate__()
        return state

    def __setstate__(self, state):
        """
        Set the state of the SubInput by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(SubInputRun, self).__setstate__(state)

        if not hasattr(self, '_source'):
            self._source = None

    def __str__(self):
        """
        Get a string version for the SubInput

        :return: the string version
        :rtype: str
        """
        if self.source_output is not None:
            return '<SubInputRun: {} => {}>'.format(self.fullid, self.source_output.fullid)
        else:
            return '<SubInputRun: {} => None>'.format(self.fullid)

    def cardinality(self, key=None, job_data=None):
        """
        Get the cardinality for this SubInput. The cardinality for a SubInputs
        is defined by the incoming link.

        :param key: key for a specific sample, can be sample index or id
        :type key: :py:class:`SampleIndex <fastr.core.sampleidlist.SampleIndex>` or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """
        if self.source is not None:
            return self.source[0].cardinality(index=key)
        else:
            return 0

    @property
    def description(self):
        return self.parent.description

    @property
    def dimensions(self):
        """
        The sample size of the SubInput
        """
        if self.source is None:
            return ()
        else:
            return self.source[0].dimensions

    @property
    def item_index(self):
        index = self.parent.index(self)
        return index

    @property
    def fullid(self):
        """
        The full defining ID for the SubInput
        """
        return '{}/{}'.format(self.parent.fullid, self.item_index)

    @property
    def input_group(self):
        """
        The id of the :py:class:`InputGroup <fastr.planning.node.InputGroup>` this
        SubInputs parent belongs to.
        """
        return self.parent.input_group

    @property
    def node(self):
        """
        The Node to which this SubInputs parent belongs
        """
        return self.parent.node

    @property
    def source_output(self):
        """
        The :py:class:`Output <fastr.planning.inputoutput.Output>`
        linked to this SubInput
        """
        if self.source is not None and len(self.source) > 0:
            return self.source[0].source
        else:
            return None

    @property
    def source(self):
        """
        A list with the source :py:class:`Link <fastr.planning.link.Link>`.
        The list is to be compatible with :py:class:`Input <fastr.planning.inputoutput.Input>`
        """
        if self._source is None:
            self.parent.remove(self)
            return []

        return [self._source]

    @source.setter
    def source(self, value):
        """
        Set new source, make sure previous link to source is released
        """
        if value is self._source:
            return

        if self._source is not None:
            self._source.destroy()

        if value is None:
            self.parent.remove(self)

        self._source = value

    def get_sourced_nodes(self):
        """
        Get a list of all :py:class:`Nodes <fastr.planning.node.Node>` connected as sources to this SubInput

        :return: list of all connected :py:class:`Nodes <fastr.planning.node.Node>`
        :rtype: list
        """
        return [x.source.node for x in self.source]

    def get_sourced_outputs(self):
        """
        Get a list of all :py:class:`Outputs <fastr.planning.inputoutput.Output>` connected as sources to this SubInput

        :return: list of all connected :py:class:`Outputs <fastr.planning.inputoutput.Output>`
        :rtype: list
        """
        return [x.source for x in self.source]

    def iteritems(self):
        """
        Iterate over the :py:class:`SampleItems <fastr.core.sampleidlist.SampleItem>`
        that are in the SubInput.

        :return: iterator yielding :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` objects
        """
        for item in self.source.items():
            yield item

    def itersubinputs(self):
        """Iterate over SubInputs (for a SubInput it will yield self and stop iterating after that)

        :return: iterator yielding  :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`

        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        yield self

    def _update(self, key, forward=True, backward=False):
        """Update the validity of the SubInput and propagate the update downstream.
        A SubInput is valid if:

        * the source Link is set and valid (see :py:meth:`Link.update <fastr.planning.link.Link.update>`)
        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        valid = True
        messages = []
        if len(self.source) == 0:
            self.parent.remove(self)
            valid = False
            messages.append('No source in this SubInput, removing!')
        elif not self.source[0].valid:
            valid = False
            messages.append('SubInput source ({}) is not valid'.format(self.source[0].id))
            messages.extend(self.source[0].messages)

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update downstream
        self.parent.update(key, forward, backward)


class OutputRun(BaseOutput, ContainsSamples):
    """
    Class representing an output of a node. It holds the output values of
    the tool ran. Output fields can be connected to inputs of other nodes.
    """

    def __init__(self, node_run, template):
        """Instantiate an Output

        :param node: the parent node the output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the output.
        :return: created Output
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``fastr.types``
        """
        self._suboutputlist = {}
        self._samples = None
        super().__init__(node_run, template.description)
        self._datatype = template.datatype
        self._listeners = []
        self._preferred_types = template.preferred_types
        self._samples = SampleCollection(template.dimnames, self)

    def __str__(self):
        """
        Get a string version for the Output

        :return: the string version
        :rtype: str
        """
        return '<OutputRun: {})>'.format(self.fullid)

    def __getitem__(self, key):
        """
        Retrieve an item from this Output. The returned value depends on what type of key used:

        * Retrieving data using index tuple: [index_tuple]
        * Retrieving data sample_id str: [SampleId]
        * Retrieving a list of data using SampleId list: [sample_id1, ..., sample_idN]
        * Retrieving a :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>` using an int or slice: [n] or [n:m]

        :param key: the key of the requested item, can be a number, slice, sample
                    index tuple or a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :type key: int, slice, :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` or tuple
        :return: the return value depends on the requested key. If the key was
                 an int or slice the corresponding :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>`
                 will be returned (and created if needed). If the key was a
                 :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                 or sample index tuple, the corresponding
                 :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned. If the
                 key was a list of :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` a tuple
                 of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned.
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>` or
                :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` or
                list of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>`
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the parent Node has not been executed
        """
        if isinstance(key, (SampleId, SampleIndex, tuple, list)):
            # If the key is a index, then get the sample id
            if isinstance(key, (SampleId, SampleIndex)):
                return self.samples[key]
            else:
                # A list or tuple of SampleId/SampleIndex
                if not all(isinstance(k, (SampleId, SampleIndex)) for k in key):
                    message = ('If a list/tuple of keys is used, all elements should be of SampleId or SampleIndex type'
                               ' found key {}'.format(key))
                    log.error(message)
                    raise exceptions.FastrValueError(message)

                return tuple(self.samples[k] for k in key)
        elif isinstance(key, (int, slice)):
            return self._get_suboutput(key)
        else:
            raise exceptions.FastrTypeError('Key should be an integer/slice (for getting a SubOutput) or an index tuple/sample_id str for getting value(s)')

    def __setitem__(self, key, value):
        """
        Store an item in the Output

        :param key: key of the value to store
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param value: the value to store
        :return: None
        :raises FastrTypeError: if key is not of correct type
        """
        if isinstance(value, SampleItem):
            self.samples[key] = value
        else:
            if not isinstance(value, (tuple, list)):
                value = (value,)

            self.samples[key] = tuple(self._cast_to_storetype(x) for x in value)

    def __getstate__(self):
        """
        Retrieve the state of the Output

        :return: the state of the object
        :rtype dict:
        """
        state = super(OutputRun, self).__getstate__()

        # Add specific fields to the state
        state['suboutputs'] = [x.__getstate__() for x in self._suboutputlist.values()]
        if self._preferred_types is not None:
            state['preferred_types'] = [x.id for x in self._preferred_types]
        else:
            state['preferred_types'] = None

        return state

    def __setstate__(self, state):
        """
        Set the state of the Output by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(OutputRun, self).__setstate__(state)

        if state['preferred_types'] is not None:
            self._preferred_types = [types[x] for x in state['preferred_types']]
        else:
            self._preferred_types = None

        suboutputlist = []
        for substate in state['suboutputs']:
            suboutput = SubOutputRun(self, slice(None))
            suboutput.__setstate__(substate)
            suboutputlist.append((suboutput.indexrep, suboutput))

        # Re-create the dict from the array
        self._suboutputlist = dict(suboutputlist)
        self._listeners = []

    def _cast_to_storetype(self, value):
        """
        Cast a given value to a DataType that matches this Outputs datatype.

        :param value: value to cast
        :return: cast value
        :rtype: DataType matching self.datatype
        """
        if isinstance(value, self.datatype):
            return value

        log.info('CAST VALUE: [{}] {!r} / {}'.format(type(value).__name__,
                                                           value,
                                                           value))

        storetype = types.match_types(self.datatype, type(value))

        if storetype is None:
            storetype = types.match_types(self.datatype)

        if not isinstance(value, storetype):
            if isinstance(value, DataType):
                log.warning('Changing value type from {} to {}'.format(type(value), storetype))
            value = storetype(str(value))

        return value

    def _get_suboutput(self, key):
        """
        Get a suboutput based on the key

        :param int, slice key: The key of the suboutput
        :return: the suboutput
        """
        # Get a string representation of the key
        if isinstance(key, slice):
            keystr = '{}:{}'.format(key.start, key.stop)
            keystr = keystr.replace('None', '')

            if key.step is not None and key.step != 1:
                keystr = '{}:{}'.format(keystr, key.step)
        else:
            keystr = str(key)

        if keystr in self._suboutputlist:
            # Re-use the same SubOutput
            subout = self._suboutputlist[keystr]
        else:
            # Create the desired SubOutput object
            subout = SubOutputRun(self, key)
            self._suboutputlist[keystr] = subout

        return subout

    def cardinality(self, key=None, job_data=None):
        """
        Cardinality of this Output, may depend on the inputs of the parent Node.

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        :raises FastrCardinalityError: if cardinality references an invalid :py:class:`Input <fastr.planning.inputoutput.Input>`
        :raises FastrTypeError: if the referenced cardinality values type cannot be case to int
        :raises FastrValueError: if the referenced cardinality value cannot be case to int
        """
        return self.cardinality_spec.calculate_execution_cardinality(key)

    @property
    def datatype(self):
        """
        The datatype of this Output
        """
        return self._datatype

    @property
    def fullid(self):
        """
        The full defining ID for the Output
        """
        if self.node is not None:
            return '{}/outputs/{}'.format(self.node.fullid, self.id)
        else:
            return 'fastr://ORPHANED/outputs/{}'.format(self.id)

    @property
    def listeners(self):
        """
        The list of :py:class:`Links <fastr.planning.link.Link>` connected to this Output.
        """
        return self._listeners

    @property
    def preferred_types(self):
        """
        The list of preferred :py:class:`DataTypes <fastr.plugins.managers.datatypemanager.DataType>`
        for this Output.
        """
        if self._preferred_types is not None and len(self._preferred_types) > 0:
            return self._preferred_types
        elif self.node.parent is not None and self.node.parent.preferred_types is not None and len(self.node.parent.preferred_types) > 0:
            return self.node.parent.preferred_types
        else:
            return types.preferred_types

    @property
    def samples(self):
        """
        The SampleCollection of the samples in this Output. None if the NodeRun
        has not yet been executed. Otherwise a SampleCollection.
        """
        return self._samples

    @property
    def valid(self):
        """
        Check if the output is valid, i.e. has a valid cardinality
        """
        return self.check_cardinality()

    @property
    def resulting_datatype(self):
        """
        The :py:class:`DataType <fastr.plugins.managers.datatypemanager.DataType>` that
        will the results of this Output will have.
        """
        requested_types = [l.target.datatype for l in self.listeners if l.target is not None]
        requested_types.append(self.datatype)

        if self.preferred_types is not None and len(self.preferred_types) > 0:
            return types.match_types(requested_types,
                                        preferred=self.preferred_types)
        else:
            return types.match_types(requested_types)

    @property
    def automatic(self):
        """
        Flag indicating that the Output is generated automatically
        without being specified on the command line
        """
        return self._description.automatic

    def iterconvergingindices(self, collapse_dims):
        """
        Iterate over all data, but collapse certain dimension to create lists
        of data.

        :param collapse_dims: dimension to collapse
        :type collapse_dims: iterable of int
        :return: iterator SampleIndex (possibly containing slices)
        """
        if all(-self.ndims <= x < self.ndims for x in collapse_dims):
            iter_dims = [range(s) for s in self.size]
            for idx in collapse_dims:
                iter_dims[idx] = slice(None),

            for idx in itertools.product(*iter_dims):
                yield SampleIndex(idx)
        else:
            raise exceptions.FastrIndexError('Index of a converging dimension {} out out of range (number of dimensions {})'.format(collapse_dims, self.ndims))

    def _update(self, key, forward=True, backward=False):
        """Update the status and validity of the Output and propagate the update the NodeRun.
        An Output is valid if:

        * the parent NodeRun is valid (see :py:meth:`NodeRun.update <fastr.planning.node.NodeRun.update>`)
        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))

        self.node.update(key, forward, backward)

        if self.node.valid:
            self._status['valid'] = True
        else:
            self._status['valid'] = False
            self._status['messages'] = ['Parent NodeRun is not valid']


class SubOutputRun(OutputRun):
    """
    The SubOutput is an Output that represents a slice of another Output.
    """

    def __init__(self, output, index):
        """Instantiate a SubOutput

        :param output: the parent output the suboutput slices.
        :param index: the way to slice the parent output
        :type index: int or slice
        :return: created SubOutput
        :raises FastrTypeError: if the output argument is not an instance of :py:class:`Output <fastr.planning.inputoutput.Output>`
        :raises FastrTypeError: if the index argument is not an ``int`` or ``slice``
        """
        if not isinstance(output, OutputRun):
            raise exceptions.FastrTypeError('Second argument for a SubOutput init should be an Output')

        if not isinstance(index, (int, slice)):
            raise exceptions.FastrTypeError('SubOutput index should be an integer or a slice, found ({}, type {})'.format(index, type(index).__name__))

        self._suboutputlist = {}  # SubOutput can have SubOutputs again
        BaseOutput.__init__(self, output.node, output.description)
        self.parent = output
        self.index = index

    def __str__(self):
        """
        Get a string version for the SubOutput

        :return: the string version
        :rtype: str
        """
        return '<SubOutputRun {}>'.format(self.fullid)

    def __getitem__(self, key):
        """
        Retrieve an item from this SubOutput. The returned value depends on what type of key used:

        * Retrieving data using index tuple: [index_tuple]
        * Retrieving data sample_id str: [SampleId]
        * Retrieving a list of data using SampleId list: [sample_id1, ..., sample_idN]
        * Retrieving a :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>` using an int or slice: [n] or [n:m]

        :param key: the key of the requested item, can be a number, slice, sample
                    index tuple or a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :type key: int, slice, :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` or tuple
        :return: the return value depends on the requested key. If the key was
                 an int or slice the corresponding :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>`
                 will be returned (and created if needed). If the key was a
                 :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                 or sample index tuple, the corresponding
                 :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned. If the
                 key was a list of :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` a tuple
                 of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned.
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>` or
                :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` or
                list of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>`
        :raises FastrTypeError: if key is not of a valid type
        """

        if isinstance(key, (int, slice)):
            return self._get_suboutput(key)

        item = self.parent[key]

        if isinstance(item, SampleItem):
            if isinstance(self.index, int):
                return SampleItem(item.index, item.id, {0: (item.data.sequence_part()[self.index],)}, item.jobs, item.failed_annotations)
            else:
                return SampleItem(item.index, item.id, {0: item.data.sequence_part()[self.index]}, item.jobs, item.failed_annotations)
        else:
            if isinstance(self.index, int):
                return tuple(SampleItem(x.index, x.id, {0: (x.data.sequence_part()[self.index],)}, x.jobs, x.failed_annotations) for x in item)
            else:
                return tuple(SampleItem(x.index, x.id, {0: x.data.sequence_part()[self.index]}, x.jobs, x.failed_annotations) for x in item)

    def __setitem__(self, key, value):
        """
        A function blocking the assignment operator. Values cannot be assigned to a SubOutput.

        :raises FastrNotImplementedError: if called
        """
        raise exceptions.FastrNotImplementedError('[{}] Cannot assign values to a SubOutput, assign to parent Output instead!'.format(self.fullid))

    def __getstate__(self):
        """
        Retrieve the state of the SubOutput

        :return: the state of the object
        :rtype dict:
        """
        state = BaseOutput.__getstate__(self)
        state['index'] = self.indexrep
        return state

    def __setstate__(self, state):
        """
        Set the state of the SubOutput by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        if isinstance(state['index'], str):
            index = [int(x) if len(x) > 0 else None for x in state['index'].split(':')]
            state['index'] = slice(*index)

        state = BaseOutput.__getstate__(self, state)

    def __len__(self):
        """Return the length of the Output.

        .. note::

            In a SubOutput this is always 1.
        """
        return 1

    @property
    def indexrep(self):
        """
        Simple representation of the index.
        """
        if isinstance(self.index, slice):
            index = '{}:{}'.format(self.index.start, self.index.stop)
            index = index.replace('None', '')

            if self.index.step is not None and self.index.step != 1:
                index = '{}:{}'.format(index, self.index.step)
        else:
            index = self.index

        return index

    def cardinality(self, key=None, job_data=None):
        """
        Cardinality of this SubOutput depends on the parent Output and ``self.index``

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        :raises FastrCardinalityError: if cardinality references an invalid :py:class:`Input <fastr.planning.inputoutput.Input>`
        :raises FastrTypeError: if the referenced cardinality values type cannot be case to int
        :raises FastrValueError: if the referenced cardinality value cannot be case to int
        """
        parent_cardinality = self.parent.cardinality(key)

        if parent_cardinality is not None:
            if isinstance(parent_cardinality, int):
                if isinstance(self.index, int):
                    if parent_cardinality >= 1:
                        return 1
                    else:
                        return 0
                else:
                    # Calculate the slice effect on a list of length parent cardinality
                    ind_range = self.index.indices(parent_cardinality)
                    return (ind_range[1] - ind_range[0]) // ind_range[2]
            else:
                return parent_cardinality
        else:
            return None

    @property
    def datatype(self):
        """
        The datatype of this SubOutput
        """
        return self.parent.datatype

    @property
    def fullid(self):
        """
        The full defining ID for the SubOutput
        """
        return '{}/{}'.format(self.parent.fullid, self.indexrep)

    @property
    def listeners(self):
        """
        The list of :py:class:`Links <fastr.planning.link.Link>` connected to this Output.
        """
        return self.parent.listeners

    @property
    def node(self):
        """
        The NodeRun to which this SubOutput belongs
        """
        return self.parent.node

    @property
    def preferred_types(self):
        """
        The list of preferred :py:class:`DataTypes <fastr.plugins.managers.datatypemanager.DataType>`
        for this SubOutput.
        """
        return self.parent.preferred_types

    @property
    def samples(self):
        """
        The :py:class:`SampleCollection <fastr.core.sampleidlist.SampleCollection>`
        for this SubOutput
        """
        return self.parent.samples

    @property
    def resulting_datatype(self):
        """
        The :py:class:`DataType <fastr.plugins.managers.datatypemanager.DataType>` that
        will the results of this SubOutput will have.
        """
        return self.parent.resulting_datatype

    def _update(self, key, forward=True, backward=False):
        """Update the status and validity of the SubOutput and propagate the update downstream.
        An SubOutput is valid if:

        * the parent NodeRun is valid (see :py:meth:`NodeRun.update <fastr.planning.node.NodeRun.update>`)

        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        self.parent.update(key, forward, backward)

        if self.node.valid:
            self._status['valid'] = True
        else:
            self._status['valid'] = False
            self._status['messages'] = ['Parent NodeRun is not valid']


class AdvancedFlowOutputRun(OutputRun):
    pass


class MacroOutputRun(OutputRun):
    @property
    def dimensions(self):
        return self.node.get_output_info(self)


class SourceOutputRun(OutputRun):
    """
    Output for a SourceNodeRun, this type of Output determines the cardinality in
    a different way than a normal NodeRun.
    """
    def __init__(self, node_run, template):
        """Instantiate a FlowOutput

        :param node: the parent node the output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the output.
        :return: created FlowOutput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``fastr.types``
        """
        super(SourceOutputRun, self).__init__(node_run, template)

        self._linearized = None

    def __getitem__(self, item):
        """
        Retrieve an item from this Output. The returned value depends on what type of key used:

        * Retrieving data using index tuple: [index_tuple]
        * Retrieving data sample_id str: [SampleId]
        * Retrieving a list of data using SampleId list: [sample_id1, ..., sample_idN]
        * Retrieving a :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>` using an int or slice: [n] or [n:m]

        :param key: the key of the requested item, can be a number, slice, sample
                    index tuple or a :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :type key: int, slice, :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` or tuple
        :return: the return value depends on the requested key. If the key was
                 an int or slice the corresponding :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>`
                 will be returned (and created if needed). If the key was a
                 :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                 or sample index tuple, the corresponding
                 :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned. If the
                 key was a list of :py:class:`SampleId <fastr.core.sampleidlist.SampleId>` a tuple
                 of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` will be returned.
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>` or
                :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` or
                list of :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>`
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the parent NodeRun has not been executed
        """
        if len(item) != 1:
            log.debug('Non-linear access to SourceOutput attempted! (linearized data: {})'.format(self.linearized))
            raise exceptions.FastrIndexError('SourceOutput only allows for linear indices')

        return self.linearized[item[0]]

    def __setitem__(self, key, value):
        """
        Store an item in the Output

        :param key: key of the value to store
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param value: the value to store
        :return: None
        :raises FastrTypeError: if key is not of correct type
        """
        super(SourceOutputRun, self).__setitem__(key, value)
        self._linearized = None

    @property
    def dimensions(self):
        """
        The dimensions of this SourceOutputRun
        """
        if self.node.nodegroup is not None:
            name = self.node.nodegroup
        else:
            name = self.node.id

        return Dimension(name, len(self.linearized)),

    @property
    def size(self):
        """
        The sample size of the SourceOutput
        """
        return len(self.linearized),

    @property
    def ndims(self):
        """
        The number of dimensions in this SourceOutput
        """
        return 1

    @property
    def linearized(self):
        """
        A linearized version of the sample data, this is lazily cached
        linearized version of the underlying SampleCollection.
        """
        if self._linearized is None:
            self._linearized = tuple(self.samples[x] for x in self.samples)

        return self._linearized

    def cardinality(self, key=None, job_data=None):
        """
        Cardinality of this SourceOutput, may depend on the inputs of the parent NodeRun.

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """
        if key is None:
            cardinalities = set()
            for sample_id in self.samples:
                cardinalities.add(self.samples[sample_id].cardinality)
            if len(cardinalities) == 1:
                return cardinalities.pop()
            else:
                return None

        if self.samples is None:
            return sympy.symbols('N_{}'.format(self.node.id.replace(' ', '_')))

        try:
            value = self[key]
        except (KeyError, IndexError):
            return None

        return len(value.data)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the API for fastr that users should use. This API will be
considered stable between major versions. If users only interact via this API
(and refrain from operating on ``parent`` attributes), their code should be
compatible within major version of fastr.
"""

import datetime
import pathlib
from typing import Any, Dict, Generic, Iterator, List, Mapping, MutableMapping, Optional, Tuple, TypeVar, Union
import os

from .. import exceptions, MISSING
from ..abc.serializable import load, save
from ..planning.network import Network as CoreNetwork
from ..planning.inputoutput import BaseInput, BaseOutput
from ..core.resourcelimit import ResourceLimit
from ..core.tool import Tool
from ..core.version import Version
from ..datatypes import BaseDataType
from ..execution.networkrun import NetworkRun


# Expose only the main functions to users, rest will follow from there
__all__ = [
    'create_network',
    'create_network_copy',
    'ResourceLimit',
]


CollapseType = Tuple[Union[int, str], ...]

VersionType = Union[Version, str]

DataTypeType = Union[BaseDataType, str]
ToolType = Union[
    Tool,
    str,
]

# The data structure used for source data, base structure can be dict of list or dict
SourceElementType = Union[str, int, float, bool]
SourceValueType = Union[Tuple[SourceElementType, ...], SourceElementType]
SourceDataType = Dict[str, Union[
    # List of values (str or tuple)
    List[SourceValueType],
    # Dict of str (id) -> value (str or tuple)
    Mapping[str, SourceValueType],
    SourceValueType,
]]

# The data structure used for sink data
SinkDataType = Union[str, Dict[str, str]]

# Different data structures that describe a state
NetworkStateType = Union['Network', CoreNetwork, dict]


T = TypeVar("T")


class SubObjectMap(Mapping[str, T], Generic[T]):
    """
    Generic object to retrieve wrapped objects from a collection
    """
    __slots__ = ('_parent', '_attribute', '_type')

    def __init__(self, parent: Any, attribute: str, type_: type):
        self._parent = parent
        self._attribute = attribute
        self._type = type_

    def __repr__(self) -> str:
        return '<{} map, items: {}>'.format(
            self._type.__name__,
            list(self.collection.keys())
        )

    def __iter__(self) -> Iterator[T]:
        for key in self.collection.__iter__():
            yield key

    def __len__(self) -> int:
        return len(self.collection)

    def __getitem__(self, item) -> T:
        return self._type(self.collection[item])

    @property
    def collection(self) -> MutableMapping[str, T]:
        return getattr(self._parent.parent, self._attribute)


class BaseWrapper:
    """
    Generic base class for wrapping fastr internal objects in a user-exposed
    API objects.
    """
    __slots__ = ('_parent',)

    def __init__(self, parent: Any):
        self._parent = parent

    def __repr__(self) -> str:
        return repr(self.parent)

    def __str__(self) -> str:
        return str(self.parent)

    @property
    def id(self) -> str:
        """
        The unique id describing this resource
        """
        return self.parent.id

    @property
    def parent(self) -> Any:
        """
        The parent object for this wrapper. This point to a non-api object
        in the internals of fastr and should not be used by normal users.
        """
        return self._parent


def cast_basewrapper(value):
    """
    Cast a value to non have BaseWrappers. BaseWrappers will be replaced by their parent.
    In Tuples each element is cast if needed.
    """
    if isinstance(value, BaseWrapper):
        value = value.parent
    elif isinstance(value, tuple):
        value = tuple(x.parent if isinstance(x, BaseWrapper) else x for x in value)
    return value


class Output(BaseWrapper):
    """
    Representation of an Output of a Node
    """
    def __getitem__(self, item: Union[int, slice]) -> 'Output':
        """
        Get a SubOuput of this Ouput. The SubOutput selects some data from the
        parent Output based on an index or slice of the cardinalty.

        :param item: the key of the requested item, can be an index or slice
        :return: the requested SubOutput with a view of the data in this Output
        """
        return Output(self.parent.__getitem__(item))


# The data structure that can be used for the source of linking
LinkSourceType = Union[Output, BaseOutput, list, dict, tuple]


class Link(BaseWrapper):
    """
    Representation of a link for editing the Network
    """
    @property
    def collapse(self) -> CollapseType:
        """
        The dimensions which the link will collapse into the cardinality
        """
        return self.parent.collapse

    @collapse.setter
    def collapse(self, value: CollapseType):
        self.parent.collapse = value

    @property
    def expand(self) -> bool:
        """
        Flag that indicates if the Link will expand the cardinality into a new
        dimension.
        """
        return self.parent.expand

    @expand.setter
    def expand(self, value: bool):
        self.parent.expand = value


class Input(BaseWrapper):
    """
    Representation of an Input of a Node
    """

    def __getitem__(self, key: str) -> 'NamedSubInput':
        if not isinstance(key, str):
            raise ValueError('Can only manipulate named sub inputs directly')

        return NamedSubInput(self.parent[key])

    def __setitem__(self, key: str, value):
        if not isinstance(key, str):
            raise ValueError('Can only manipulate named sub inputs directly')

        self.parent[key] = cast_basewrapper(value)

    def append(self, value: LinkSourceType) -> Link:
        """
        Create a link from give resource to a new SubInput.

        :param value: The source for the link to be created
        :return: The newly created link
        """
        return self.parent.append(cast_basewrapper(value))

    @property
    def input_group(self) -> str:
        """
        The input group of this Input. This property can be read and changed.
        Changing the input group of an Input will influence the data flow in
        a Node (see :ref:`advanced-flow-node` for details).
        """
        return self.parent.input_group

    @input_group.setter
    def input_group(self, value: str):
        self.parent.input_group = value

    def __lshift__(self, other: LinkSourceType) -> Union[Link, Tuple[Link, ...]]:
        """
        This operator allows the easy creation of Links to this Input using the ``<<`` operator.
        Creating links can be done by:

        .. code-block:: python

            # Generic form
            >> link = input << output
            >> link = input << ['some', 'data']  # Create a constant node

            # Examples
            >> link1 = addint.inputs['left_hand'] << source1.input
            >> link2 = addint.inputs['right_hand'] << [1, 2, 3]

            # Mutliple links
            >> links = addints.inputs['left_hand'] << (source1.output, source2.output, source3.output)

        The last example would return a tuple with three links.

        :param other: the target to create the link from, this can be an Output, a tuple of Outputs, or a data
                      structure that can be used as the data for a ConstantNode
        :return: Newly created link(s)
        """
        # Make sure wrappers are cast to non-wrappers
        other = cast_basewrapper(other)

        return self.parent << other

    def __rrshift__(self, other: LinkSourceType) -> Union[Link, Tuple[Link, ...]]:
        """
        This operator allows to use the ``>>`` operator as alternative to using the ``<<`` operator.
        See the :py:meth:`__lshift__ operator <fastr.api.Input.__lshift__>` for details.

        :param other: the target to create the link from
        :return: Newly created link(s)
        """
        # Make sure wrappers are cast to non-wrappers
        other = cast_basewrapper(other)

        return other >> self.parent


class NamedSubInput(BaseWrapper):
    """
    A named sub-input. This is only used in cases where a tool accepts a mapping
    as an input. For example, if a command-line tool would accept::

        --set key1=value1 key2=value2

    Or something similar, where key1 and key2 are arbitrary strings. This can
    be translated in fastr to::

        tool.inputs['set']['key1'] << value1
        tool.inputs['set']['key2'] << value2

    """
    def __lshift__(self, other: LinkSourceType) -> Union[Link, Tuple[Link, ...]]:
        """
        This operator allows the easy creation of Links to this Input using the ``<<`` operator.
        Creating links can be done by:

        .. code-block:: python

            # Generic form
            >> link = input << output
            >> link = input << ['some', 'data']  # Create a constant node

            # Examples
            >> link1 = addint.inputs['left_hand'] << source1.input
            >> link2 = addint.inputs['right_hand'] << [1, 2, 3]

            # Mutliple links
            >> links = addints.inputs['left_hand'] << (source1.output, source2.output, source3.output)

        The last example would return a tuple with three links.

        :param other: the target to create the link from, this can be an Output, a tuple of Outputs, or a data
                      structure that can be used as the data for a ConstantNode
        :return: Newly created link(s)
        """
        # Make sure wrappers are cast to non-wrappers
        other = cast_basewrapper(other)

        return self.parent << other

    def __rrshift__(self, other: LinkSourceType) -> Union[Link, Tuple[Link, ...]]:
        """
        This operator allows to use the ``>>`` operator as alternative to using the ``<<`` operator.
        See the :py:meth:`__lshift__ operator <fastr.api.Input.__lshift__>` for details.

        :param other: the target to create the link from
        :return: Newly created link(s)
        """
        # Make sure wrappers are cast to non-wrappers
        other = cast_basewrapper(other)

        return other >> self.parent

    def append(self, value: LinkSourceType) -> Link:
        """
        Create a link from give resource to a new SubInput.

        :param value: The source for the link to be created
        :return: The newly created link
        """
        return self.parent.append(cast_basewrapper(value))


class Node(BaseWrapper):
    """
    Representation of Node for editing the Network
    """
    __slots__ = ('_inputs', '_outputs')

    def __init__(self, parent):
        super().__init__(parent)
        self._inputs = InputMap(self, 'inputs', Input)
        self._outputs = SubObjectMap(self, 'outputs', Output)

    @property
    def inputs(self) -> 'InputMap':
        """
        Mapping object containing all Inputs of a Node
        """
        return self._inputs

    @property
    def outputs(self) -> SubObjectMap[Output]:
        """
        Mapping object containing all Outputs of a Node
        """
        return self._outputs

    @property
    def input(self) -> Input:
        """
        In case there is only a single Inputs in a Node, this can be used as a short hand.
        In that case it is basically the same as ``list(node.inputs.values()[0])``.
        """
        if len(self.inputs) == 1:
            return next(iter(self.inputs.values()))
        else:
            raise KeyError('There is not 1 unique input, cannot use short-cut!')

    @input.setter
    def input(self, value):
        if len(self.inputs) == 1:
            input = next(iter(self.inputs.values()))
            input << value
        else:
            raise KeyError('There is not 1 unique input, cannot use short-cut!')

    @property
    def output(self) -> Output:
        """
        In case there is only a single Outputs in a Node, this can be used as a short hand.
        In that case it is basically the same as ``list(node.outputs.values()[0])``.
        """
        if len(self.outputs) == 1:
            return next(iter(self.outputs.values()))
        else:
            raise KeyError('There is not 1 unique outputs, cannot use short-cut!')

    @property
    def merge_dimensions(self) -> Union[str, Tuple[str, ...]]:
        return self._parent.merge_dimensions

    @merge_dimensions.setter
    def merge_dimensions(self, value: Union[str, Tuple[str, ...]]):
        self._parent.merge_dimensions = value


class InputMap(SubObjectMap[Input]):
    def __setitem__(self, key, value):
        # Make sure wrappers are cast to non-wrappers
        value = cast_basewrapper(value)
        self.collection[key] = value


class Network(BaseWrapper):
    """
    Representation of a Network for the creating and adapting Networks
    """
    __slots__ = ('_node_map',)

    def __init__(self, id, version=None):
        parent = CoreNetwork(id_=id, version=version)
        super().__init__(parent)
        self._node_map = SubObjectMap(self, 'nodelist', Node)

    @property
    def nodes(self) -> SubObjectMap[Node]:
        return self._node_map

    @property
    def version(self) -> Version:
        """
        Version of the Network (so users can keep track of their version)
        """
        return self.parent.version

    def create_node(self,
                    tool: ToolType,
                    tool_version: str,
                    id: str = None,
                    step_id: str = None,
                    resources: ResourceLimit = None,
                    node_group: str = None) -> Node:
        """
        Create a Node in this Network. The Node will be automatically added to
        the Network.

        :param tool: The Tool to base the Node on in the form: ``name/space/toolname:version``
        :param tool_version: The version of the tool wrapper to use
        :param id: The id of the node to be created
        :param step_id: The step to add the created node to
        :param resources: The resources required to run this node
        :param node_group: The group the node belongs to, this can be
                           important for FlowNodes and such, as they
                           will have matching dimension names.
        :return: the newly created node
        """

        if not isinstance(tool, (str, Tool)):
            raise exceptions.FastrTypeError('The tool argument should be either a Tool or a str')

        resources = resources or ResourceLimit()
        return Node(self.parent.create_node(
            tool=tool,
            tool_version=tool_version,
            id_=id,
            stepid=step_id,
            resources=resources.copy(),
            nodegroup=node_group
        ))

    def create_macro(self,
                     network: Union[NetworkStateType, Tool, str],
                     id: str = None) -> Node:
        """
        Create macro node (a node which actually contains a network used as node
        inside another network).

        :param network: The network to use, this can be a network (state), a
                        macro tool, or the path to a python file that contains
                        a function create_network which returns the desired
                        network.
        :param id: The id of the node to be created
        :return: the newly created node
        """
        return Node(self.parent.create_macro(network=network, id_=id))

    def create_constant(self,
                        datatype: DataTypeType,
                        data: SourceDataType,
                        id: str = None,
                        step_id: str = None,
                        resources: ResourceLimit = None,
                        node_group: str = None) -> Node:
        """
        Create a ConstantNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the constant node
        :param data: The data to hold in the constant node
        :param id: The id of the constant node to be created
        :param step_id: The step to add the created constant node to
        :param resources: The resources required to run this node
        :param node_group: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :return: the newly created constant node
        """
        resources = resources or ResourceLimit()
        return Node(self.parent.create_constant(
            datatype=datatype,
            data=data,
            id_=id,
            stepid=step_id,
            resources=resources.copy(),
            nodegroup=node_group
        ))

    def create_link(self,
                    source: Union[Input, BaseInput],
                    target: Union[Output, BaseOutput],
                    id: str = None,
                    collapse: CollapseType = None,
                    expand: bool = False) -> Link:
        """
        Create a link between two Nodes and add it to the current Network.

        :param source: the output that is the source of the link
        :param target: the input that is the target of the link
        :param id: the id of the link
        :param collapse: The dimensions to collapse in this link.
        :param expand: Flag to expand cardinality into a new dimension
        :return: the created link
        """
        if isinstance(source, Output):
            source = source.parent

        if isinstance(target, Input):
            target = target.parent

        return Link(self.parent.create_link(
            source=source,
            target=target,
            id_=id,
            collapse=collapse,
            expand=expand
        ))

    def create_source(self,
                      datatype: DataTypeType,
                      id: str = None,
                      step_id: str = None,
                      resources: ResourceLimit = None,
                      node_group: str = None) -> Node:
        """
        Create a SourceNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the source source_node
        :type datatype: :py:class:`BaseDataType <fastr.plugins.managers.datatypemanager.BaseDataType>`
        :param str id: The id of the source source_node to be created
        :param str step_id: The step to add the created source source_node to
        :param resources: The resources required to run this node
        :param str node_group: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :return: the newly created source source_node
        :rtype: :py:class:`SourceNode <fastr.core.source_node.SourceNode>`
        """
        resources = resources or ResourceLimit()
        return Node(self.parent.create_source(
            datatype=datatype,
            id_=id,
            stepid=step_id,
            resources=resources.copy(),
            nodegroup=node_group
        ))

    def create_sink(self,
                    datatype: DataTypeType,
                    id: str = None,
                    step_id: str = None,
                    resources: ResourceLimit = None,
                    node_group: str = None) -> Node:
        """
        Create a SinkNode in this Network. The Node will be automatically added to
        the Network.

        :param datatype: The DataType of the sink node
        :param id: The id of the sink node to be created
        :param step_id: The step to add the created sink node to
        :param resources: The resources required to run this node
        :param str node_group: The group the node belongs to, this can be
                              important for FlowNodes and such, as they
                              will have matching dimension names.
        :return: the newly created sink node
        """
        resources = resources or ResourceLimit()
        return Node(self.parent.create_sink(
            datatype=datatype,
            id_=id,
            stepid=step_id,
            resources=resources.copy(),
            nodegroup=node_group,
        ))

    def create_reference(self,
                         source_data: SourceDataType,
                         output_directory: str,
                         ):
        """
        Create reference data to test a Network against

        :param source_data: The source data to use for the reference
        :param output_directory: The directory to store the reference
        """
        self._parent.create_reference(source_data=source_data,
                                      output_directory=output_directory)

    def draw(self,
             file_path: str = None,
             draw_dimensions: bool = True,
             hide_unconnected: bool = True,
             expand_macros: Union[bool, int] = 1,
             font_size: int = 14) -> Optional[str]:
        """
        Draw a graphical representation of the Network

        :param str file_path: The path of the file to create, the extension will control the image type
        :param bool draw_dimensions: Flag to control if the dimension sizes should be drawn
                                     in the figure, default is true
        :param bool expand_macros: Flag to control if and how macro nodes should be expanded,
                                   by default 1 level is expanded
        :return: path of the image created or None if failed
        """
        if file_path is not None:
            file_path, ext = os.path.splitext(file_path)
        else:
            file_path = self.id
            ext = 'svg'

        if not ext:
            ext = 'svg'

        return self.parent.draw_network(name=file_path,
                                        img_format=ext.lstrip('.'),
                                        draw_dimension=draw_dimensions,
                                        hide_unconnected=hide_unconnected,
                                        expand_macro=expand_macros,
                                        font_size=font_size)

    def _validate_source_value(self, value: Any) -> bool:
        if not isinstance(value, tuple):
            value = value,

        return all(isinstance(x, (int, float, bool, str)) or x is MISSING for x in value)

    def _validate_source_data(self, data: Any) -> bool:
        if isinstance(data, Mapping):
            # Validate keys to be either str or Tuple[str, ...]
            for key in data.keys():
                if not isinstance(key, str):
                    return False

            return all(self._validate_source_value(x) for x in data.values())
        elif isinstance(data, list):
            return all(self._validate_source_value(x) for x in data)
        elif isinstance(data, (str, float, int, bool)):
            return True
        else:
            return False

    def execute(self,
                source_data: SourceDataType,
                sink_data: SinkDataType,
                tmpdir: str = None,
                timestamp: Union[datetime.datetime, str] = None,
                blocking: bool = True,
                execution_plugin: Optional[str] = None,
                tracking_id: Optional[str] = None) -> NetworkRun:
        """
        Execute the network with the given source and sink data.

        :param source_data: Source data to use as an input
        :param sink_data: Sink rules to use for determining the outputs
        :param tmpdir: The scratch directory to use for this network run, if
                       an existing directory is given, fastr will try to resume
                       a network run (see :ref:`continuing-network`)
        :param timestamp: The timestamp of the network run (useful for retrying
                          or continuing previous runs)
        :param blocking: Flag to indicate if the execution should be blocking
                         or launched in a background thread
        :param execution_plugin: The execution plugin to use for this run
        :param tracking_id: The tracking id for this run, to be able to track this run from the logs
        :return: The network run object for the started execution
        """
        for source_id in self.parent.sourcelist.keys():
            if not self._validate_source_data(source_data.get(source_id)):
                raise exceptions.FastrValueError(f'Invalid source data for source "{source_id}", found: {source_data.get(source_id)}')

        return self.parent.execute(
                sourcedata=source_data,
                sinkdata=sink_data,
                tmpdir=tmpdir,
                timestamp=timestamp,
                blocking=blocking,
                execution_plugin=execution_plugin,
                tracking_id=tracking_id,
            )


    def dependencies(self):
        """
        Returns a full list of the dependencies of the network.

        :return: Return a list of dict containing:
                 'node_id': ID of the node using the tool
                 'namespace': Namespace of the tool
                 'tool_id': Tool ID
                 'command_version': Command Version 
                 'tool_version': Tool Version
        """
        return self.parent.dependencies()

    @classmethod
    def load(cls,
             filename: Union[str, pathlib.Path]) -> 'Network':
        """
        Load Network from a YAML file

        :param str filename:
        :return: loaded network
        :rtype: Network
        """
        result = BaseWrapper.__new__(Network)

        # Load Network and create correct node map
        result._parent = load(filename)
        result._node_map = SubObjectMap(result, 'nodelist', Node)
        return result

    def save(self,
             filename: Union[str, pathlib.Path]):
        """
        Save the Network to a YAML file

        :param filename: Path of the file to save to
        """
        save(self.parent, filename)

    def is_valid(self):
        return self._parent.is_valid()


def create_network(id: str,
                   version: VersionType = None) -> Network:
    """
    Create a new Network object

    :param id: id of the network
    :param version: version of the network
    :return:
    """
    return Network(id=id, version=version)


def create_network_copy(network_state: NetworkStateType) -> Network:
    """
    Create a network based on another Network state. The network state can be a Network
    or the state gotten from a Network with __getstate__.

    :param network_state: Network (state) to create a copy of
    :return: The rebuilt network
    """
    if isinstance(network_state, Network):
        network_state = network_state.parent

    if isinstance(network_state, CoreNetwork):
        network_state = network_state.__getstate__()

    # Create the copy of the Network
    result = BaseWrapper.__new__(Network)
    result._parent = CoreNetwork.deserialize(network_state)
    result._node_map = SubObjectMap(result.parent, 'nodelist', Node)

    return result



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from fastr.utils.cmd import add_parser_doc_link
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


def get_parser():
    parser = argparse.ArgumentParser(
        description="Verify fastr resources, at the moment only tool definitions"
                    "are supported."
    )
    parser.add_argument('resource_type', metavar='TYPE', choices=['tool'],
                        help='Type of resource to verify (e.g. tool)')
    parser.add_argument('path', help='path of the resource to verify')
    parser.add_argument('--createtest', '-c', action='store_true', 
                        help='Create a reference result for a tool test')
    return parser


def main():
    """
    Verify fastr resources, at the moment only tool definitions are supported.
    """
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    import fastr
    from fastr.utils.verify import verify_tool, create_tool_test

    resource_type = args.resource_type.lower()
    if resource_type == 'tool':
        if args.createtest:
            create_tool_test(args.path)
        else:
            verify_tool(args.path)
    else:
        fastr.log.error('Unknown resource type {}'.format(resource_type))



import os
import gzip
import shutil
from tempfile import mkdtemp

import fastr
from ..abc.serializable import Serializable, ReadWriteHandler
from ..core.tool import Tool
from ..helpers import iohelpers, config
from .. import exceptions


def verify_resource_loading(filename: str, log=fastr.log):
    """
    Verify that a resource file can be loaded. Returns loaded object.

    :param filename: path of the object to load
    :param log: the logger to use to send messages to
    :return: loaded resource
    """
    name, ext = os.path.splitext(filename)

    # Check if file is gzipped
    if ext == '.gz':
        compressed = True
        name, ext = os.path.splitext(filename)
    else:
        compressed = False

    # Read file data
    log.info('Trying to read file with compression {}'.format('ON' if compressed else 'OFF'))
    if compressed:
        try:
            with gzip.open(filename, 'r') as file_handle:
                data = file_handle.read()
        except:
            log.error('Problem reading gzipped file: {}'.format(filename))
            return None
    else:
        try:
            with open(filename, 'r') as file_handle:
                data = file_handle.read()
        except:
            log.error('Problem reading normal file: {}'.format(filename))
            return None

    log.info('Read data from file successfully')

    # Try to read tool doc based on serializer matching the extension
    serializer = ext[1:]
    log.info('Trying to load file using serializer "{}"'.format(serializer))

    try:
        serializer = ReadWriteHandler.get_handler(serializer)
    except KeyError:
        log.error('No matching serializer found for "{}"'.format(serializer))
        return None

    load_func = serializer.loads

    try:
        doc = load_func(data)
    except Exception as exception:
        log.error('Could not load data using serializer "{}", encountered exception: {}'.format(serializer,
                                                                                                      exception))
        return None

    return doc


def verify_tool_schema(doc, log=fastr.log):
    """
    Verify the tool schema. Returns checked loaded object.

    :param doc: loaded object to check
    :param log: the logger to use to send messages to
    :return: object with checked schema
    """
    # Match the data to the schema for Tools
    log.info('Validating data against Tool schema')
    serializer = Tool.get_serializer()

    try:
        doc = serializer.instantiate(doc)
    except Exception as exception:
        log.error('Encountered a problem when verifying the Tool schema: {}'.format(exception))
        return None
    return doc


def verify_tool_instantiate(doc, filename, log=fastr.log):
    """
    Verify the tool schema. Returns checked loaded object.

    :param doc: loaded object
    :param filename: filename of the tool definition
    :param log: the logger to use to send messages to
    :return: Tool object
    """
    # Create the Tool object as the final test
    log.info(f'Instantiating Tool object')
    try:
        tool = Tool(doc)
        tool.filename = filename
    except Exception as exception:
        log.error('Encountered a problem when creating the Tool object: {}'.format(exception))
    return tool


def verify_tool(filename, log=fastr.log, perform_tests=True):
    """
    Verify that a tool correctly works. Returns Tool.

    :param filename: filename of the tool definition
    :param log: the logger to use to send messages to
    :param perform_test: Boolean to 
    :return: Tool object
    """
    # Load the file
    doc = verify_resource_loading(filename, log)

    if not doc:
        log.error('Could not load data successfully from  {}'.format(filename))
        return None

    # Match the data to the schema for Tools
    doc = verify_tool_schema(doc, log)

    # Create the Tool object as the final test
    tool = verify_tool_instantiate(doc, filename, log)

    if perform_tests:
        log.info('Testing tool...')
        try:
            tool.test()
        except fastr.exceptions.FastrValueError as e:
            log.error('Tool is not valid: {}'.format(e))

    return tool


def create_tool_test(filename, log=fastr.log):
    """
    Create test for fastr verify tool.

    By running `fastr verify -c tool FILENAME` the input data in the folders 
    under 'tests' in the tool definition is processed by the tool. The 
    output data is written to a folder in each test folder. In each test folder
    a gzipped pickle is created which is used to verify the working of the tool
    at a later time.

    :param filename: filename of the tool definition
    :param log: the logger to use to send messages to
    """
    # Load the file
    doc = verify_resource_loading(filename, log)

    if not doc:
        log.error('Could not load data successfully from  {}'.format(filename))
        return None

    doc = verify_tool_schema(doc, log)

    tool = verify_tool_instantiate(doc, filename, log)

    log.info('Loaded tool {} successfully'.format(tool))
    tool_dir = os.path.dirname(tool.filename)
    for test in tool.tests:
        reference_data_dir = os.path.abspath(os.path.join(tool_dir, test))
        try:
            if not isinstance(reference_data_dir, str):
                raise exceptions.FastrTypeError('reference_data_dir should be a string!')

            if reference_data_dir.startswith('vfs://'):
                reference_data_dir = vfs_plugin.url_to_path(reference_data_dir)

            if not os.path.isdir(reference_data_dir):
                raise exceptions.FastrTypeError('The reference_data_dir should be pointing to an existing directory!'
                                                ' {} does not exist'.format(reference_data_dir))

            test_data = iohelpers.load_json(
                os.path.join(reference_data_dir, tool.TOOL_REFERENCE_FILE_NAME)
            )

            input_data = {}

            for key, value in test_data['input_data'].items():
                if not isinstance(value, (tuple, list)):
                    value = value,

                # Set the $REFDIR correctly (the avoid problems with moving the reference dir)
                value = tuple(x.replace('$REFDIR', reference_data_dir) if isinstance(x, str) else x for x in value)
                input_data[key] = value

            temp_results_dir = None
            try:
                # Create temporary output directory
                temp_results_dir = os.path.normpath(mkdtemp(
                    prefix='fastr_tool_test_{}_'.format(tool.id), dir=config.mounts['tmp']
                ))

                # Create a new reference for comparison
                log.info('Creating new reference data for comparison...')
                try:
                    if not os.path.exists(os.path.join(reference_data_dir, 
                                                       tool.TOOL_RESULT_FILE_NAME)):
                        # Copy original __fastr_tool_ref__.json, 
                        # so it doesn't get overwritten. Afterwards move it back.
                        shutil.copy(os.path.join(reference_data_dir, 
                                                 tool.TOOL_REFERENCE_FILE_NAME), 
                                    os.path.join(reference_data_dir, 
                                                 ''.join([tool.TOOL_REFERENCE_FILE_NAME, 
                                                          'bak'])))
                        tool.create_reference(input_data,
                                              reference_data_dir,
                                              mount_name='__ref_tmp__',
                                              copy_input=False,
                                              input_datatypes=test_data['input_datatypes'])
                        log.info(
                            'Reference result for testing the {}/{} created in {}.'.format( 
                                 tool.ns_id, 
                                 tool.command_version,
                                 reference_data_dir
                        ))
                        shutil.copy(os.path.join(reference_data_dir, 
                                                 tool.TOOL_REFERENCE_FILE_NAME), 
                                    os.path.join(reference_data_dir, 
                                                 ''.join(['__output', 
                                                          tool.TOOL_REFERENCE_FILE_NAME])))
                        shutil.move(os.path.join(reference_data_dir, 
                                                 ''.join([tool.TOOL_REFERENCE_FILE_NAME, 
                                                          'bak'])),
                                    os.path.join(reference_data_dir, 
                                                 tool.TOOL_REFERENCE_FILE_NAME))
                    else:
                        log.warning(
                            'Reference result for testing the {}/{} tool already exists in {}!'.format( 
                                 tool.ns_id, 
                                 tool.command_version,
                                 reference_data_dir
                        ))
                except Exception as exception:
                    log.warning('Encountered exception when trying to run the {}/{} tool!'.format(
                        tool.ns_id, tool.command_version)
                    )
                    log.warning('Exception: [{}] {}'.format(type(exception).__name__, exception))
            finally:
                # Clean up
                log.info('Removing temp result directory {}'.format(temp_results_dir))
                if temp_results_dir is not None and os.path.isdir(temp_results_dir):
                    shutil.rmtree(temp_results_dir, ignore_errors=True)
        except exceptions.FastrTypeError:
            message = 'Reference data in {} is not valid!'.format(reference_data_dir)
            log.warning(message)
    return


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The plugins module holds all plugins loaded by Fastr. It is empty on start
and gets filled by the BasePluginManager
"""

from ..helpers import config

from .managers.pluginmanager import PluginManager
from .managers.executionpluginmanager import ExecutionPluginManager
from .managers.interfacemanager import InterfacePluginManager
from .managers.iopluginmanager import IOPluginManager
from .managers.targetmanager import TargetManager

from .managers.toolmanager import ToolManager
from .managers.networkmanager import NetworkManager

from .. import resources

# Main plugin manager
plugins = PluginManager()

# All the submanagers for specific access
execution_plugins = ExecutionPluginManager(plugins)
interfaces = InterfacePluginManager(plugins)
ioplugins = IOPluginManager(plugins)
targets = TargetManager(plugins)

# Create the toollist
tools = ToolManager(config.tools_path)

# Creat the network list
networklist = NetworkManager(config.networks_path)

# Register resources
resources.plugins = plugins
resources.execution_plugins = execution_plugins
resources.interfaces = interfaces
resources.ioplugins = ioplugins
resources.targets = targets

resources.tools = tools
resources.networks = networklist
